from cmdbox.app import common, web
from cmdbox.app.commons import convert
from fastapi import FastAPI, Request, Response, HTTPException
from pathlib import Path
from starlette.datastructures import UploadFile
from starlette.responses import PlainTextResponse
from starlette.middleware.sessions import SessionMiddleware
from typing import Any, Dict, List
from uvicorn.config import Config
import asyncio
import cv2
import datetime
import gevent
import logging
import sys
import traceback
import time


class Web(web.Web):
    def __init__(self, logger:logging.Logger, data:Path, appcls=None, ver=None,
                 redis_host:str="localhost", redis_port:int=6379, redis_password:str=None, svname:str='server',
                 client_only:bool=False, doc_root:Path=None, gui_html:str=None, filer_html:str=None, result_html:str=None, users_html:str=None,
                 audit_html:str=None, assets:List[str]=None, signin_html:str=None, signin_file:str=None, gui_mode:bool=False,
                 web_features_packages:List[str]=None, web_features_prefix:List[str]=[],
                 showimg_html:str=None, webcap_html:str=None, anno_html:str=None):
        """
        iinferクライアント側のwebapiサービス

        Args:
            logger (logging): ロガー
            data (Path): コマンドやパイプラインの設定ファイルを保存するディレクトリ
            appcls ([type], optional): アプリケーションクラス. Defaults to None.
            ver ([type], optional): バージョン. Defaults to None.
            redis_host (str, optional): Redisサーバーのホスト名. Defaults to "localhost".
            redis_port (int, optional): Redisサーバーのポート番号. Defaults to 6379.
            redis_password (str, optional): Redisサーバーのパスワード. Defaults to None.
            svname (str, optional): 推論サーバーのサービス名. Defaults to 'server'.
            client_only (bool, optional): クライアントのみのサービスかどうか. Defaults to False.
            doc_root (Path, optional): カスタムファイルのドキュメントルート. フォルダ指定のカスタムファイルのパスから、doc_rootのパスを除去したパスでURLマッピングします。Defaults to None.
            gui_html (str, optional): GUIのHTMLファイル. Defaults to None.
            filer_html (str, optional): ファイラーのHTMLファイル. Defaults to None.
            result_html (str, optional): 結果表示のHTMLファイル. Defaults to None.
            users_html (str, optional): ユーザー管理のHTMLファイル. Defaults to None.
            audit_html (str, optional): 監査ログのHTMLファイル. Defaults to None.
            assets (List[str], optional): 静的ファイルのリスト. Defaults to None.
            signin_html (str, optional): ログイン画面のHTMLファイル. Defaults to None.
            signin_file (str, optional): ログイン情報のファイル. Defaults to args.signin_file.
            gui_mode (bool, optional): GUIモードかどうか. Defaults to False.
            web_features_packages (List[str], optional): webfeatureのパッケージ名のリスト. Defaults to None.
            web_features_prefix (List[str], optional): webfeatureのプレフィックスのリスト. Defaults to None.
            showimg_html (str, optional): 画像表示のHTMLファイル. Defaults to None.
            webcap_html (str, optional): ウェブカメラのHTMLファイル. Defaults to None.
            anno_html (str, optional): アノテーション画面のHTMLファイル. Defaults to None.
        """
        doc_root = doc_root if doc_root is not None else Path(__file__).parent.parent / 'web'
        assets = assets if assets is not None else [str(Path(__file__).parent.parent / 'web' / 'assets')]
        super().__init__(logger=logger, data=data, appcls=appcls, ver=ver,
                         redis_host=redis_host, redis_port=redis_port, redis_password=redis_password, svname=svname,
                         client_only=client_only, doc_root=doc_root, gui_html=gui_html, filer_html=filer_html,
                         result_html=result_html, users_html=users_html, audit_html=audit_html,
                         assets=assets, signin_html=signin_html, signin_file=signin_file, gui_mode=gui_mode,
                         web_features_packages=web_features_packages, web_features_prefix=web_features_prefix)
        self.showimg_html = Path(showimg_html) if showimg_html is not None else Path(__file__).parent.parent / 'web' / 'showimg.html'
        self.webcap_html = Path(webcap_html) if webcap_html is not None else Path(__file__).parent.parent / 'web' / 'webcap.html'
        self.anno_html = Path(anno_html) if anno_html is not None else Path(__file__).parent.parent / 'web' / 'annotation.html'

        self.showimg_html_data = None
        self.webcap_html_data = None
        self.anno_html_data = None
        if self.logger.level == logging.DEBUG:
            self.logger.debug(f"web init parameter: showimg_html={self.showimg_html}")
            self.logger.debug(f"web init parameter: webcap_html={self.webcap_html}")
            self.logger.debug(f"web init parameter: anno_html={self.anno_html}")

    def webcap(self, allow_host:str="0.0.0.0", listen_webcap_port:int=8082,
               image_type:str='capture', outputs_key:List[str]=None, capture_frame_width:int=None, capture_frame_height:int=None,
               capture_count:int=5, capture_fps:int=5):
        """
        Webキャプチャを起動する

        Args:
            allow_host (str, optional): 許可ホスト. Defaults to "
            listen_webcap_port (int, optional): リスンポート. Defaults to 8082.
            image_type (str, optional): 画像タイプ. Defaults to 'capture'.
            outputs_key (list, optional): 出力キー. Defaults to None.
            capture_frame_width (int, optional): キャプチャフレーム幅. Defaults to None.
            capture_frame_height (int, optional): キャプチャフレーム高さ. Defaults to None.
            capture_count (int, optional): キャプチャ回数. Defaults to 5.
            capture_fps (int, optional): キャプチャFPS. Defaults to 5.
        """
        self.allow_host = allow_host
        self.listen_webcap_port = listen_webcap_port
        self.image_type = image_type
        self.outputs_key = outputs_key
        self.capture_frame_width = capture_frame_width
        self.capture_frame_height = capture_frame_height
        self.capture_count = capture_count
        self.capture_fps = capture_fps
        self.count = 0
        if self.logger.level == logging.DEBUG:
            self.logger.debug(f"web webcap parameter: allow_host={self.allow_host}")
            self.logger.debug(f"web webcap parameter: listen_webcap_port={self.listen_webcap_port}")
            self.logger.debug(f"web webcap parameter: image_type={self.image_type}")
            self.logger.debug(f"web webcap parameter: outputs_key={self.outputs_key}")
            self.logger.debug(f"web webcap parameter: capture_frame_width={self.capture_frame_width}")
            self.logger.debug(f"web webcap parameter: capture_frame_height={self.capture_frame_height}")
            self.logger.debug(f"web webcap parameter: capture_count={self.capture_count}")
            self.logger.debug(f"web webcap parameter: capture_fps={self.capture_fps}")

        app = FastAPI()
        app.add_middleware(SessionMiddleware, secret_key=common.random_string())
        self.init_webfeatures(app)

        @app.get('/webcap/pub_img', response_class=PlainTextResponse)
        async def pub_img_webcap_chk(req:Request, res:Response):
            return "ok"

        @app.post('/webcap/pub_img')
        async def pub_img_webcap(req:Request, res:Response, title:str=None):
            """
            webcap(プロキシ)から送信されてきた画像を、指定された画像タイプに変換して標準出力する
            """
            if self.logger.level == logging.DEBUG:
                self.logger.debug(f"web.pub_img_webcap: headers={dict(res.headers)}")
            content_type = req.headers.get('content-type')
            if content_type is None or not content_type.startswith('multipart/form-data'):
                raise HTTPException(status_code=400, detail='Expected multipart request.')
            res.headers['Access-Control-Allow-Origin'] = '*'

            image_type = self.image_type
            try:
                tm = time.perf_counter()
                form = await req.form()
                files = {key: value for key, value in form.items() if isinstance(value, UploadFile)}
                for fn in files.keys():
                    cap:bytes = files[fn].file.read()
                    if cap is None:
                        continue
                    output_image_name = datetime.datetime.now().strftime("%Y%m%d%H%M%S%f")
                    cap = cap.decode().replace('data:image/jpeg;base64,','')
                    if self.logger.level == logging.DEBUG:
                        ret_str = common.to_str(cap, slise=100)
                        self.logger.debug(f"web.pub_img_webcap: cap={ret_str}")
                    img_npy = convert.imgbytes2npy(convert.b64str2bytes(cap))
                    if self.capture_frame_width is not None and self.capture_frame_height is not None:
                        img_npy = cv2.resize(img_npy, (self.capture_frame_width, self.capture_frame_height), interpolation=cv2.INTER_NEAREST)
                    img_b64 = None
                    if image_type == 'capture' or image_type is None:
                        image_type = 'capture'
                        img_b64 = convert.npy2b64str(img_npy)
                    else:
                        img_b64 = convert.bytes2b64str(convert.npy2imgfile(img_npy, image_type=image_type))
                    output_image_name = f"{output_image_name}.{image_type}"

                    t, b64, h, w, c, fn = image_type, img_b64, img_npy.shape[0], img_npy.shape[1], img_npy.shape[2] if len(img_npy.shape) > 2 else -1, output_image_name
                    ret = f"{t},"+b64+f",{h},{w},{c},{fn}"
                    if self.logger.level == logging.DEBUG:
                        ret_str = common.to_str(ret, slise=100)
                        self.logger.debug(f"web.pub_img_webcap: ret={ret_str}")
                    common.print_format(ret, False, tm, None, False)
                    tm = time.perf_counter()
                    self.count += 1
                if self.capture_count > 0 and self.count >= self.capture_count:
                    self.is_running = False
                    #gevent.sleep(3)
                    self.logger.info(f"Exit webcap. allow_host={self.allow_host} listen_webcap_port={self.listen_webcap_port}")

            except Exception as e:
                self.logger.warning('pub_img_webcap error', exc_info=True)
                return common.to_str(dict(warn=f'pub_img_webcap error. {traceback.format_exc()}'))

            ret = common.to_str(dict(success='pub_img_webcap to stdout.'))
            return ret

        self.is_running = True
        log_config = {"version":1,
                      "formatters":{
                            "fmt":{
                                "format":'%(levelname)s[%(asctime)s] - %(message)s',
                                "class":"logging.Formatter"}},
                      "handlers":{
                            "std":{
                                "class":"cmdbox.app.commons.loghandler.ColorfulStreamHandler",
                                "level":"NOTSET",
                                "formatter":"fmt",
                                "stream":"ext://sys.stdout"}},
                      "loggers":{
                            "uvicorn":{
                                "handlers":["std"],
                                "level":"NOTSET",
                                "qualname":"uvicorn"}}}

        class ThreadedASGI(web.ThreadedASGI):
            def __init__(self, app:FastAPI, logger:logging.Logger, config:Config, gunicorn_config:Dict[str, Any]=None):
                super().__init__(app, logger, config, gunicorn_config=gunicorn_config, force_single=True)

        th = ThreadedASGI(app, self.logger, config=Config(app=app, host=self.allow_host, port=self.listen_webcap_port))
        th.start()
        try:
            tm = time.time()
            while self.is_running:
                if time.time() - tm > 15:
                    tm = time.time()
                    sys.stderr.write(f"webcap running. count={self.count}, capture_count={self.capture_count}\n")
                gevent.sleep(1)
            th.stop()
        except KeyboardInterrupt:
            th.stop()



from cmdbox.app import common
import platform
import logging

class Redis(object):
    def __init__(self, logger: logging.Logger, wsl_name: str = None, wsl_user: str = None):
        self.logger = logger
        self.wsl_name = wsl_name
        self.wsl_user = wsl_user

    def docker_run(self, port: int, password: str):
        if port is None is None:
            return {"warn":f"port option is required."}
        if password is None:
            return {"warn":f"password option is required."}
        docker_cmd = str(f"docker run -itd --name redis --rm -e TZ=UTC -p {port}:{port} -e REDIS_PASSWORD={password} ubuntu/redis:latest "
                         f"bash -c '"
                         f"sed -i -e \"s/^bind 127.0.0.1/bind 0.0.0.0/\" /etc/redis/redis.conf;"
                         f"sed -i -e \"s/^# maxmemory <bytes>/maxmemory 5gb/\" /etc/redis/redis.conf;"
                         f"sed -i -e \"s/^# maxmemory-policy noeviction/maxmemory-policy volatile-lru/\" /etc/redis/redis.conf;"
                         f"sed -i -e \"s/^# proto-max-bulk-len 512mb/proto-max-bulk-len 5gb/\" /etc/redis/redis.conf;"
                         f"redis-server /etc/redis/redis.conf --requirepass {password};" # --daemonize yes 
                         f"bash'")
        if platform.system() == 'Windows':
            if self.wsl_name is None:
                return {"warn":f"wsl_name option is required."}
            if self.wsl_user is None:
                return {"warn":f"wsl_user option is required."}
            code, _, _cmd = common.cmd(f"wsl -d {self.wsl_name} -u {self.wsl_user} {docker_cmd}" ,self.logger)
            return {"output":code}
        elif platform.system() == 'Linux':
            code, _, _cmd = common.cmd(docker_cmd, self.logger)
            return {"output":code}
        else:
            return {"warn":f"Unsupported platform."}

    def docker_stop(self):
        if platform.system() == 'Windows':
            if self.wsl_name is None:
                return {"warn":f"wsl_name option is required."}
            if self.wsl_user is None:
                return {"warn":f"wsl_user option is required."}
            code, _, _cmd = common.cmd(f"wsl -d {self.wsl_name} -u {self.wsl_user} docker stop redis", self.logger)
            code, _, _cmd = common.cmd(f"wsl -d {self.wsl_name} -u {self.wsl_user} docker rm redis", self.logger)
            common.cmd(f"wsl --shutdown", self.logger)
            return {"output":code}
        elif platform.system() == 'Linux':
            code, _, _cmd = common.cmd(f"docker stop redis", self.logger)
            return {"output":code}
        else:
            return {"warn":f"Unsupported platform."}



from iinfer.app.postprocesses import det_judge
from typing import List
import logging

class ClaJadge(det_judge.DetJadge):
    def __init__(self, logger:logging.Logger,
                 ok_score_th:float=None, ok_classes:List[int]=None, ok_labels:List[str]=None,
                 ng_score_th:float=None, ng_classes:List[int]=None, ng_labels:List[str]=None,
                 ext_score_th:float=None, ext_classes:List[int]=None, ext_labels:List[str]=None,
                 nodraw:bool=False, output_preview:bool=False):
        """
        Image Classification推論結果のクラススコア元に、この画像としてOK/NG/Grayの判定結果を追加する後処理クラスです。

        Args:
            logger (logging.Logger): ロガー
            ok_score_th (float): ok判定確定のスコアの閾値
            ok_classes (List[int]): ok判定確定のクラスのリスト
            ok_labels (List[str]): ok判定確定のラベルのリスト
            ng_score_th (float): ng判定確定のスコアの閾値
            ng_classes (List[int]): ng判定確定のクラスのリスト
            ng_labels (List[str]): ng判定確定のラベルのリスト
            ext_score_th (float): gray判定確定のスコアの閾値
            ext_classes (List[int]): gray判定確定のクラスのリスト
            ext_labels (List[str]): gray判定確定のラベルのリスト
            nodraw (bool): 描画しない
            output_preview (bool): プレビューを出力する
        """
        super().__init__(logger, ok_score_th=ok_score_th, ok_classes=ok_classes, ok_labels=ok_labels,
                         ng_score_th=ng_score_th, ng_classes=ng_classes, ng_labels=ng_labels,
                         ext_score_th=ext_score_th, ext_classes=ext_classes, ext_labels=ext_labels,
                         nodraw=nodraw, output_preview=output_preview)




from cmdbox.app import client
from cmdbox.app.commons import convert
from pathlib import Path
from iinfer.app import common as cmn
from iinfer.app.commons import module
from typing import List
import base64
import cv2
import datetime
import glob
import logging
import json
import numpy as np
import time


class Client(client.Client):
    def __init__(self, logger:logging.Logger, redis_host:str = "localhost", redis_port:int = 6379, redis_password:str = None, svname:str = 'server'):
        """
        Redisサーバーとの通信を行うクラス

        Args:
            logger (logging): ロガー
            redis_host (str, optional): Redisサーバーのホスト名. Defaults to "localhost".
            redis_port (int, optional): Redisサーバーのポート番号. Defaults to 6379.
            redis_password (str, optional): Redisサーバーのパスワード. Defaults to None.
            svname (str, optional): 推論サーバーのサービス名. Defaults to 'server'.
        """
        super().__init__(logger, redis_host, redis_port, redis_password, svname)

    def deploy(self, name:str, model_img_width:int, model_img_height:int, model_file:str, model_conf_file:List[Path], predict_type:str,
               custom_predict_py:Path, label_file:Path, color_file:Path,
               before_injection_conf:Path, before_injection_type:List[str], before_injection_py:List[Path],
               after_injection_conf:Path, after_injection_type:List[str], after_injection_py:List[Path],
               train_dataset:Path, train_dataset_upload:bool, train_type:str, custom_train_py:Path,
               overwrite:bool, retry_count:int=3, retry_interval:int=5, timeout:int = 60):
        """
        モデルをRedisサーバーにデプロイする

        Args:
            name (str): モデル名
            model_img_width (int): 画像の幅
            model_img_height (int): 画像の高さ
            model_file (str): モデルファイルのパス又はURL
            model_conf_file (List[Path]): モデル設定ファイルのパス
            predict_type (str): 推論方法のタイプ
            custom_predict_py (Path): 推論スクリプトのパス
            label_file (Path): ラベルファイルのパス
            color_file (Path): 色ファイルのパス
            before_injection_conf (Path): 推論前処理設定ファイルのパス
            before_injection_type (List[str]): 推論前処理タイプ
            before_injection_py (List[Path]): 推論前処理スクリプトのパス
            after_injection_type (List[str]): 推論後処理タイプ
            after_injection_conf (Path): 推論後処理設定ファイルのパス
            after_injection_py (List[Path]): 推論後処理スクリプトのパス
            train_dataset (Path): 学習データセットディレクトリのパス
            train_dataset_upload (bool): 学習データセットをサーバーにアップロードするかどうか
            train_type (str): 学習方法のタイプ
            custom_train_py (Path): 学習スクリプトのパス
            overwrite (bool): モデルを上書きするかどうか
            retry_count (int, optional): リトライ回数. Defaults to 3.
            retry_interval (int, optional): リトライ間隔. Defaults to 5.
            timeout (int, optional): タイムアウト時間. Defaults to 60.

        Returns:
            dict: Redisサーバーからの応答
        """
        if name is None or name == "":
            self.logger.warning(f"name is empty.")
            return {"error": f"name is empty."}
        if " " in name:
            self.logger.warning(f"name contains whitespace.")
            return {"error": f"name contains whitespace."}
        if model_file is None:
            self.logger.warning(f"model_file or model_file is empty.")
            return {"error": f"model_file or model_file is empty."}
        if predict_type is None:
            self.logger.warning(f"predict_type is empty.")
            return {"error": f"predict_type is empty."}
        if predict_type not in cmn.BASE_MODELS and predict_type != "Custom":
            self.logger.warning(f"Unknown predict_type. {predict_type}")
            return {"error": f"Unknown predict_type. {predict_type}"}
        if predict_type == 'Custom':
            if custom_predict_py is None or not custom_predict_py.exists():
                self.logger.warning(f"custom_predict_py path {str(custom_predict_py)} does not exist")
                return {"error": f"custom_predict_py path {str(custom_predict_py)} does not exist"}
            with open(custom_predict_py, "rb") as pf:
                custom_predict_py_b64 = base64.b64encode(pf.read()).decode('utf-8')
            pred = module.load_custom_predict(custom_predict_py, self.logger)
            model_img_width = pred.IMAGE_WIDTH
            model_img_height = pred.IMAGE_HEIGHT
        else:
            custom_predict_py_b64 = None
            if model_img_width is None or model_img_width <= 0:
                model_img_width = cmn.BASE_MODELS[predict_type]['image_width']
            if model_img_height is None or model_img_height <= 0:
                model_img_height = cmn.BASE_MODELS[predict_type]['image_height']
        if before_injection_type is not None and len(before_injection_type) > 0:
            for t in before_injection_type:
                if t not in cmn.BASE_BREFORE_INJECTIONS:
                    self.logger.warning(f"Unknown before_injection_type. {t}")
                    return {"error": f"Unknown before_injection_type. {t}"}
            before_injection_type = ','.join(before_injection_type)
        if after_injection_type is not None and len(after_injection_type) > 0:
            for t in after_injection_type:
                if t not in cmn.BASE_AFTER_INJECTIONS:
                    self.logger.warning(f"Unknown after_injection_type. {t}")
                    return {"error": f"Unknown after_injection_type. {t}"}
            after_injection_type = ','.join(after_injection_type)
        def _conf_b64(name:str, conf:Path):
            if conf is not None and not conf.exists():
                self.logger.warning(f"{name} {conf} does not exist")
                return False, {"error": f"{name} {conf} does not exist"}
            elif conf is not None:
                with open(conf, "rb") as lf:
                    conf_b64 = base64.b64encode(lf.read()).decode('utf-8')
            else:
                conf_b64 = None
            return True, conf_b64
        ret, label_file_b64 = _conf_b64("label_file", label_file)
        if not ret: return label_file_b64
        ret, color_file_b64 = _conf_b64("color_file", color_file)
        if not ret: return color_file_b64
        ret, before_injection_conf_b64 = _conf_b64("before_injection_conf", before_injection_conf)
        if not ret: return before_injection_conf_b64
        ret, after_injection_conf_b64 = _conf_b64("after_injection_conf", after_injection_conf)
        if not ret: return after_injection_conf_b64
        if model_file is not None and not model_file.startswith("http://") and not model_file.startswith("https://"):
            model_file = Path(model_file)
            if model_file.exists():
                with open(model_file, "rb") as mf:
                    model_bytes_b64 = base64.b64encode(mf.read()).decode('utf-8')
            elif predict_type != 'Custom' and cmn.BASE_MODELS[predict_type]['required_model_weight']:
                self.logger.warning(f"model_file {model_file} does not exist")
                return {"error": f"model_file {model_file} does not exist"}
            else:
                model_bytes_b64 = None
        else:
            model_bytes_b64 = None
        def _name_b64(aname:str, files:List[Path]):
            if files is not None:
                b64s = []
                names = []
                for p in files:
                    if not p.exists():
                        self.logger.warning(f"{aname} {p} does not exist")
                        return False, {"error": f"{aname} {p} does not exist"}, None
                    with open(p, "rb") as f:
                        b64s.append(base64.b64encode(f.read()).decode('utf-8'))
                    names.append(p.name)
                return True, ','.join(names), ','.join(b64s)
            return True, None, None
        ret, model_conf_file_name, model_conf_bytes_b64 = _name_b64("model_conf_file", model_conf_file)
        if not ret: return model_conf_file_name
        ret, before_injection_py_name, before_injection_py_b64 = _name_b64("before_injection_py", before_injection_py)
        if not ret: return before_injection_py_name
        ret, after_injection_py_name, after_injection_py_b64 = _name_b64("after_injection_py", after_injection_py)
        if not ret: return after_injection_py_name

        if train_dataset is not None:
            if train_type is None:
                self.logger.warning(f"train_type is empty.Required if train_dataset is specified.")
                return {"error": f"train_type is empty.Required if train_dataset is specified."}
            if not train_dataset.exists():
                self.logger.warning(f"train_dataset path {str(train_dataset)} does not exist")
                return {"error": f"train_dataset path {str(train_dataset)} does not exist"}
            if not train_dataset.is_dir():
                self.logger.warning(f"train_dataset path {str(train_dataset)} does not directory")
                return {"error": f"train_dataset path {str(train_dataset)} does not directory"}
        if train_type is not None and train_dataset is None:
            self.logger.warning(f"train_dataset is empty.Required if train_type is specified.")
            return {"error": f"train_dataset is empty.Required if train_type is specified."}
        if train_type is not None and train_type not in cmn.BASE_TRAIN_MODELS and train_type != "Custom":
            self.logger.warning(f"Unknown train_type. {train_type}")
            return {"error": f"Unknown train_type. {train_type}"}
        if train_type == 'Custom':
            if custom_train_py is None or not custom_train_py.exists():
                self.logger.warning(f"custom_train_py path {str(custom_train_py)} does not exist")
                return {"error": f"custom_train_py path {str(custom_train_py)} does not exist"}
            with open(custom_train_py, "rb") as pf:
                custom_train_py_b64 = base64.b64encode(pf.read()).decode('utf-8')
        else:
            custom_train_py_b64 = None

        res_json = self.redis_cli.send_cmd('deploy', [name, str(model_img_width), str(model_img_height), predict_type,
                                           model_file.name if isinstance(model_file, Path) else model_file, model_bytes_b64,
                                           model_conf_file_name, model_conf_bytes_b64,
                                           custom_predict_py_b64, label_file_b64, color_file_b64,
                                           before_injection_conf_b64, before_injection_type, before_injection_py_name, before_injection_py_b64,
                                           after_injection_conf_b64, after_injection_type, after_injection_py_name, after_injection_py_b64,
                                           'input' if train_dataset is not None and train_dataset_upload else None,
                                           train_type, custom_train_py_b64, overwrite],
                                           retry_count=retry_count, retry_interval=retry_interval, outstatus=False, timeout=timeout)

        if train_dataset is not None and train_dataset_upload:
            data_files = glob.glob(str(train_dataset / "**" / "*"), recursive=True)
            svpath = Path(f"/{name}")
            for f in data_files:
                file_path = Path(f)
                if file_path.is_dir():
                    continue
                p = f.replace(str(train_dataset), '', 1).replace('\\', '/')
                up_path = svpath / 'input' / p[1:] if p.startswith('/') else svpath / p
                rj = self.file_upload(up_path, file_path, scope="server", mkdir=True, orverwrite=True,
                                            retry_count=retry_count, retry_interval=retry_interval, timeout=timeout)
                if 'success' not in rj:
                    return rj

        return res_json

    def train(self, name:str, overwrite:bool, retry_count:int=3, retry_interval:int=5, timeout:int = 3*3600):
        """
        モデルをRedisサーバーで学習する

        Args:
            name (str): モデル名
            overwrite (bool): モデルを上書きするかどうか
            retry_count (int, optional): リトライ回数. Defaults to 3.
            retry_interval (int, optional): リトライ間隔. Defaults to 5.
            timeout (int, optional): タイムアウト時間. Defaults to 3*3600.

        Returns:
            dict: Redisサーバーからの応答
        """
        if name is None or name == "":
            self.logger.warning(f"name is empty.")
            return {"error": f"name is empty."}
        if " " in name:
            self.logger.warning(f"name contains whitespace.")
            return {"error": f"name contains whitespace."}

        res_json = self.redis_cli.send_cmd('train', [name, overwrite],
                                           retry_count=retry_count, retry_interval=retry_interval, outstatus=False, timeout=timeout)
        return res_json

    def deploy_list(self, retry_count:int=3, retry_interval:int=5, timeout:int = 60):
        """
        デプロイされたモデルのリストを取得する

        Args:
            retry_count (int, optional): リトライ回数. Defaults to 3.
            retry_interval (int, optional): リトライ間隔. Defaults to 5.
            timeout (int, optional): タイムアウト時間. Defaults to 60.

        Returns:
            dict: Redisサーバーからの応答
        """
        res_json = self.redis_cli.send_cmd('deploy_list', [],
                                           retry_count=retry_count, retry_interval=retry_interval, timeout=timeout)
        return res_json

    def undeploy(self, name:str, retry_count:int=3, retry_interval:int=5, timeout:int = 60):
        """
        モデルをRedisサーバーからアンデプロイする

        Args:
            name (str): モデル名
            retry_count (int, optional): リトライ回数. Defaults to 3.
            retry_interval (int, optional): リトライ間隔. Defaults to 5.
            timeout (int, optional): タイムアウト時間. Defaults to 60.

        Returns:
            dict: Redisサーバーからの応答
        """
        if name is None or name == "":
            self.logger.warning(f"name is empty.")
            return {"error": f"name is empty."}
        res_json = self.redis_cli.send_cmd('undeploy', [name],
                                           retry_count=retry_count, retry_interval=retry_interval, timeout=timeout)
        return res_json

    def start(self, name:str, model_provider:str = 'CPUExecutionProvider', use_track:bool=False, gpuid:int=None,
              retry_count:int=3, retry_interval:int=5, timeout:int = 60):
        """
        モデルをRedisサーバーで起動する

        Args:
            name (str): モデル名
            model_provider (str, optional): 推論実行時のモデルプロバイダー。デフォルトは'CPUExecutionProvider'。
            use_track (bool): Multi Object Trackerを使用するかどうか, by default False
            gpuid (int): GPU ID, by default None
            retry_count (int, optional): リトライ回数. Defaults to 3.
            retry_interval (int, optional): リトライ間隔. Defaults to 5.
            timeout (int, optional): タイムアウト時間. Defaults to 60.

        Returns:
            dict: Redisサーバーからの応答
        """
        if name is None or name == "":
            self.logger.warning(f"name is empty.")
            return {"error": f"name is empty."}
        if model_provider is None or model_provider == "":
            self.logger.warning(f"model_provider is empty.")
            return {"error": f"model_provider is empty."}
        res_json = self.redis_cli.send_cmd('start', [name, model_provider, str(use_track), str(gpuid)],
                                           retry_count=retry_count, retry_interval=retry_interval, timeout=timeout)
        return res_json

    def stop(self, name:str,
             retry_count:int=3, retry_interval:int=5, timeout:int = 60):
        """
        モデルをRedisサーバーで停止する

        Args:
            name (str): モデル名
            retry_count (int, optional): リトライ回数. Defaults to 3.
            retry_interval (int, optional): リトライ間隔. Defaults to 5.
            timeout (int, optional): タイムアウト時間. Defaults to 60.

        Returns:
            dict: Redisサーバーからの応答
        """
        if name is None or name == "":
            self.logger.warning(f"name is empty.")
            return {"error": f"name is empty."}
        res_json = self.redis_cli.send_cmd('stop', [name],
                                           retry_count=retry_count, retry_interval=retry_interval, timeout=timeout)
        return res_json


    def stop_server(self, retry_count:int=3, retry_interval:int=5, timeout:int = 60):
        """
        Redisサーバーを停止する

        Args:
            retry_count (int, optional): リトライ回数. Defaults to 3.
            retry_interval (int, optional): リトライ間隔. Defaults to 5.
            timeout (int, optional): タイムアウト時間. Defaults to 60.

        Returns:
            dict: Redisサーバーからの応答
        """
        res_json = self.redis_cli.send_cmd('stop_server', [], retry_count=retry_count, retry_interval=retry_interval, timeout=timeout)
        return res_json

    def predict(self, name:str, image = None, image_file = None, image_file_enable:bool=True, pred_input_type:str = 'jpeg',
                output_image_file:str = None, output_preview:bool=False, nodraw:bool=False,
                retry_count:int=3, retry_interval:int=5, timeout:int = 60):
        """
        画像をRedisサーバーに送信し、推論結果を取得する

        Args:
            name (str): モデル名
            image (np.ndarray | bytes, optional): 画像データ. Defaults to None. np.ndarray型の場合はデコードしない(RGBであること).
            image_file (str|file-like object, optional): 画像ファイルのパス. Defaults to None.
            image_file_enable (bool, optional): 画像ファイルを使用するかどうか. Defaults to True. image_fileがNoneでなく、このパラメーターがTrueの場合はimage_fileを使用する.
            pred_input_type (str, optional): 画像の形式. Defaults to 'jpeg'.
            output_image_file (str, optional): 予測結果の画像ファイルのパス. Defaults to None.
            output_preview (bool, optional): 予測結果の画像をプレビューするかどうか. Defaults to False.
            nodraw (bool, optional): 描画フラグ. Defaults to False.
            retry_count (int, optional): リトライ回数. Defaults to 3.
            retry_interval (int, optional): リトライ間隔. Defaults to 5.
            timeout (int, optional): タイムアウト時間. Defaults to 60.

        Returns:
            dict: Redisサーバーからの応答
        """
        spredtime = time.perf_counter()
        if name is None or name == "":
            self.logger.warning(f"name is empty.")
            return {"error": f"name is empty."}
        if image is None and image_file is None:
            self.logger.warning(f"image and image_file is empty.")
            return {"error": f"image and image_file is empty."}
        npy_b64 = None
        simgloadtime = time.perf_counter()
        if image_file is not None and image_file_enable:
            if type(image_file) == str:
                if not Path(image_file).exists():
                    self.logger.warning(f"Not found image_file. {image_file}.")
                    return {"error": f"Not found image_file. {image_file}."}
            if pred_input_type == 'jpeg' or pred_input_type == 'png' or pred_input_type == 'bmp':
                f = None
                try:
                    f = image_file if type(image_file) is not str else open(image_file, "rb")
                    img_npy = convert.imgfile2npy(f)
                finally:
                    if f is not None: f.close()
            elif pred_input_type == 'capture':
                f = None
                try:
                    f = image_file if type(image_file) is not str else open(image_file, "r", encoding='utf-8')
                    res_list = []
                    for line in f:
                        if type(line) is bytes:
                            line = line.decode('utf-8')
                        capture_data = line.strip().split(',')
                        t = capture_data[0]
                        img = capture_data[1]
                        h = int(capture_data[2])
                        w = int(capture_data[3])
                        c = int(capture_data[4])
                        fn = Path(capture_data[5].strip())
                        if t == 'capture':
                            img_npy = convert.b64str2npy(img, shape=(h, w, c) if c > 0 else (h, w))
                        else:
                            img_npy = convert.imgbytes2npy(convert.b64str2bytes(img))
                        res_json = self.predict(name, image=img_npy, image_file=fn, image_file_enable=False,
                                                output_image_file=output_image_file, output_preview=output_preview, nodraw=nodraw,
                                                retry_count=retry_count, retry_interval=retry_interval, timeout=timeout)
                        res_list.append(res_json)
                    if len(res_list) <= 0:
                        return {"warn": f"capture file is no data."}
                    elif len(res_list) == 1:
                        return res_list[0]
                    return res_list
                except UnicodeDecodeError as e:
                    self.logger.error(f"capture file or pred_input_type setting is invalid. pred_input_type={pred_input_type}. {e}", exc_info=True)
                    return {"error": f"capture file or pred_input_type setting is invalid. pred_input_type={pred_input_type}. {e}"}
                finally:
                    if f is not None: f.close()
            elif pred_input_type == 'output_json':
                f = None
                try:
                    f = image_file if type(image_file) is not str else open(image_file, "r", encoding='utf-8')
                    res_list = []
                    for line in f:
                        res_json = json.loads(line)
                        if not ("output_image" in res_json and "output_image_shape" in res_json and "output_image_name" in res_json):
                            self.logger.warn(f"image_file data is invalid. Not found output_image or output_image_shape or output_image_name key.")
                            continue
                        img_npy = convert.b64str2npy(res_json["output_image"], shape=res_json["output_image_shape"])
                        res_json = self.predict(name, image=img_npy, image_file=Path(res_json['output_image_name'].strip()), image_file_enable=False,
                                                output_image_file=output_image_file, output_preview=output_preview, nodraw=nodraw,
                                                retry_count=retry_count, retry_interval=retry_interval, timeout=timeout)
                        res_list.append(res_json)
                    if len(res_list) <= 0:
                        return {"warn": f"output_json file is no data."}
                    elif len(res_list) == 1:
                        return res_list[0]
                    return res_list
                finally:
                    if f is not None: f.close()
            else:
                self.logger.warning(f"pred_input_type is invalid. {pred_input_type}.")
                return {"error": f"pred_input_type is invalid. {pred_input_type}."}
        else:
            if type(image) == np.ndarray:
                img_npy = image
                if image_file is None: image_file = f'{datetime.datetime.now().strftime("%Y%m%d%H%M%S%f")}.capture'
                image_file_enable = False
            elif pred_input_type == 'capture':
                image = image.decode(encoding="utf-8") if type(image) is bytes else image
                capture_data = image.split(',')
                if len(capture_data) < 6:
                    self.logger.warning(f"capture data is invalid. {image}.")
                    return {"error": f"capture data is invalid. {image}."}
                t = capture_data[0]
                img = capture_data[1]
                h = int(capture_data[2])
                w = int(capture_data[3])
                c = int(capture_data[4])
                if image_file is None: image_file = capture_data[5]
                image_file_enable = False
                if t == 'capture':
                    img_npy = convert.b64str2npy(img, shape=(h, w, c) if c > 0 else (h, w))
                else:
                    img_npy = convert.imgbytes2npy(convert.b64str2bytes(img))
            elif pred_input_type == 'output_json':
                res_json = json.loads(image)
                if not ("output_image" in res_json and "output_image_shape" in res_json and "output_image_name" in res_json):
                    self.logger.warning(f"image_file data is invalid. Not found output_image or output_image_shape or output_image_name key.")
                    return {"error": f"image_file data is invalid. Not found output_image or output_image_shape or output_image_name key."}
                if res_json["output_image_name"].endswith(".capture"):
                    img_npy = convert.b64str2npy(res_json["output_image"], shape=res_json["output_image_shape"])
                else:
                    img_bytes = convert.b64str2bytes(res_json["output_image"])
                    img_npy = convert.imgbytes2npy(img_bytes)
                if image_file is None: image_file = res_json["output_image_name"]
                image_file_enable = False
            elif pred_input_type == 'jpeg' or pred_input_type == 'png' or pred_input_type == 'bmp':
                img_npy = convert.imgbytes2npy(image)
                if image_file is None: image_file = f'{datetime.datetime.now().strftime("%Y%m%d%H%M%S%f")}.{pred_input_type}'
                image_file_enable = False
            else:
                self.logger.warning(f"pred_input_type is invalid. {pred_input_type}.")
                return {"error": f"pred_input_type is invalid. {pred_input_type}."}

        eimgloadtime = time.perf_counter()
        npy_b64 = convert.npy2b64str(img_npy)
        res_json = self.redis_cli.send_cmd('predict',
                                [name, npy_b64, str(nodraw), str(img_npy.shape[0]), str(img_npy.shape[1]),
                                str(img_npy.shape[2] if len(img_npy.shape) > 2 else '-1'), image_file],
                                retry_count=retry_count, retry_interval=retry_interval, timeout=timeout)
        soutputtime = time.perf_counter()
        if "output_image" in res_json and "output_image_shape" in res_json:
            img_npy = convert.b64str2npy(res_json["output_image"], res_json["output_image_shape"])
            if output_image_file is not None:
                exp = Path(output_image_file).suffix
                exp = exp[1:] if exp[0] == '.' else exp
                convert.npy2imgfile(img_npy, output_image_file=output_image_file, image_type=exp)
            if output_preview:
                # RGB画像をBGR画像に変換
                try:
                    cv2.imshow('preview', convert.bgr2rgb(img_npy))
                    cv2.waitKey(1)
                except Exception as e:
                    self.logger.error(f"cv2.imshow error. {e}", exc_info=True)
                    pass
        eoutputtime = time.perf_counter()
        epredtime = time.perf_counter()
        if "success" in res_json:
            if "performance" not in res_json["success"]:
                res_json["success"]["performance"] = []
            performance = res_json["success"]["performance"]
            performance.append(dict(key="cl_imgload", val=f"{eimgloadtime-simgloadtime:.3f}s"))
            performance.append(dict(key="cl_output", val=f"{eoutputtime-soutputtime:.3f}s"))
            performance.append(dict(key="cl_pred", val=f"{epredtime-spredtime:.3f}s"))
        return res_json

    def read_dir(self, glob_str:str, read_input_type:str='jpeg', image_type:str='capture', root_dir:Path=Path('.'), include_hidden=True,
                 moveto:Path=None, polling:bool=False, polling_count:int=10, polling_interval:int=1):
        """
        ディレクトリ内の画像ファイルを取得します

        Args:
            glob_str (str): 読込むファイルのglobパターン
            read_input_type (str, optional): 読み込みタイプ. Defaults to 'jpeg'.
            image_type (str, optional): 画像の形式. Defaults to 'capture'.
            root_dir (Path, optional): 検索の基準となるルートディレクトリ. Defaults to Path('.').
            include_hidden (bool, optional): 隠しファイルを含めるかどうか. Defaults to True.
            moveto (Path, optional): 読み込んだファイルを移動する先のディレクトリ. Defaults to None.
            polling (bool, optional): 定期的にディレクトリ内の読込みを繰り返すかどうか. Defaults to False.
            polling_count (int, optional): ポーリング回数. Defaults to 10.
            polling_interval (int, optional): ポーリング間隔. Defaults to 1.

        Yields:
            Tuple[str, str, int, int, int, str]: 画像の形式, 画像のBase64文字列, 画像の高さ, 画像の幅, 画像の色数, 画像のファイル名
        """
        if root_dir is None or not root_dir.exists():
            self.logger.warning(f"root_dir {root_dir} does not exist.")
            raise FileNotFoundError(f"root_dir {root_dir} does not exist.")
        if polling_interval <= 0:
            self.logger.warning(f"polling_interval {polling_interval} is invalid.")
            raise ValueError(f"polling_interval {polling_interval} is invalid.")
        def _read_file(p:Path, image_type:str):
            img_npy = convert.imgfile2npy(p)
            if image_type == 'capture':
                img_b64 = convert.npy2b64str(img_npy)
            elif image_type != 'capture':
                img_b64 = convert.bytes2b64str(convert.npy2imgfile(img_npy, image_type=image_type))
            return image_type, img_b64, img_npy.shape[0], img_npy.shape[1], img_npy.shape[2] if len(img_npy.shape) > 2 else -1, p.name

        def _read_dir(glob_str:str, read_input_type:str, image_type:str, root_dir:Path, include_hidden:bool, moveto:Path):
            files = glob.glob(glob_str, root_dir=root_dir, include_hidden=include_hidden, recursive=True)
            for file in files:
                p = root_dir / file
                if not p.is_file():
                    continue
                if self.logger.level == logging.DEBUG:
                    self.logger.debug(f"read file: {p}")
                if read_input_type == 'capture':
                    with open(p, 'r', encoding='utf-8') as f:
                        for line in f:
                            cel = line.split(',')
                            if cel[0] != image_type:
                                if cel[0] == 'capture':
                                    img_npy = convert.b64str2npy(cel[1], shape=(int(cel[2]), int(cel[3]), int(cel[4]) if int(cel[4]) > 0 else 1))
                                else:
                                    img_npy = convert.imgbytes2npy(convert.b64str2bytes(cel[1]))
                                cel[0] = image_type
                                cel[1] = convert.npy2b64str(img_npy)
                            yield cel[0], cel[1], cel[2], cel[3], cel[4], cel[5]
                elif read_input_type == 'filelist':
                    with open(p, 'r', encoding='utf-8') as f:
                        for line in f:
                            p = Path(line)
                            cel = _read_file(p, image_type)
                            yield cel[0], cel[1], cel[2], cel[3], cel[4], cel[5]
                else:
                    cel = _read_file(p, image_type)
                    yield cel[0], cel[1], cel[2], cel[3], cel[4], cel[5]
                if moveto is not None:
                    mv = moveto / file
                    if self.logger.level == logging.DEBUG:
                        self.logger.debug(f"move file: {p} => {mv}")
                    mv.parent.mkdir(parents=True, exist_ok=True)
                    p.rename(mv)
        
        if polling:
            count = 0
            while True:
                for cel in _read_dir(glob_str, read_input_type, image_type, root_dir, include_hidden, moveto):
                    yield cel
                count += 1
                if count >= polling_count and polling_count > 0:
                    break
                time.sleep(polling_interval)
        else:
            for cel in _read_dir(glob_str, read_input_type, image_type, root_dir, include_hidden, moveto):
                yield cel

    def capture(self, capture_device='0', image_type:str='capture', capture_frame_width:int=None, capture_frame_height:int=None,
                capture_fps:int=1000, output_preview:bool=False):
        """
        ビデオをキャプチャしてその結果を出力する

        Args:
            capture_device (int or str): キャプチャするディバイス、ビデオデバイスのID, ビデオファイルのパス。rtspのURL. by default 0
            image_type (str, optional): 画像の形式. Defaults to 'capture'.
            capture_frame_width (int): キャプチャするビデオのフレーム幅, by default None
            capture_frame_height (int): キャプチャするビデオのフレーム高さ, by default None
            capture_fps (int): キャプチャするビデオのフレームレート, by default 1000
            output_preview (bool, optional): 予測結果の画像をプレビューするかどうか. Defaults to False.
        """
        if capture_device.isdecimal():
            capture_device = int(capture_device)
        cap = cv2.VideoCapture(capture_device)
        cap.set(cv2.CAP_PROP_BUFFERSIZE, 0)
        cap.set(cv2.CAP_PROP_FPS, 30)
        cap.set(cv2.CAP_PROP_FOURCC, cv2.VideoWriter_fourcc('H', '2', '6', '4'))
        if capture_frame_width is not None:
            cap.set(cv2.CAP_PROP_FRAME_WIDTH, capture_frame_width)
        if capture_frame_height is not None:
            cap.set(cv2.CAP_PROP_FRAME_HEIGHT, capture_frame_height)
        try:
            interval = float(1 / capture_fps)
            self.is_running = True
            while self.is_running:
                start = time.perf_counter()
                ret, frame = cap.read()
                output_image_name = datetime.datetime.now().strftime("%Y%m%d%H%M%S%f")
                if ret:
                    if capture_frame_width is not None and capture_frame_height is not None:
                        frame = cv2.resize(frame, (capture_frame_width, capture_frame_height), interpolation=cv2.INTER_NEAREST)
                    img_npy = convert.bgr2rgb(frame)
                    if output_preview:
                        # RGB画像をBGR画像に変換
                        cv2.imshow('preview', convert.bgr2rgb(img_npy))
                        cv2.waitKey(1)
                    img_b64 = None
                    if image_type == 'capture' or image_type is None:
                        image_type = 'capture'
                        img_b64 = convert.npy2b64str(img_npy)
                    else:
                        img_b64 = convert.bytes2b64str(convert.npy2imgfile(img_npy, image_type=image_type))
                    output_image_name = f"{output_image_name}.{image_type}"
                    yield image_type, img_b64, img_npy.shape[0], img_npy.shape[1], img_npy.shape[2] if len(img_npy.shape) > 2 else -1, output_image_name
                else:
                    self.logger.warning(f"Capture failed. devide_id={capture_device}", stack_info=True)
                    break
                end = time.perf_counter()
                if interval - (end - start) > 0:
                    time.sleep(interval - (end - start))

        except KeyboardInterrupt:
            self.logger.warning("KeyboardInterrupt", exc_info=True)
        finally:
            cap.release()



from cmdbox.app.features.cli import cmdbox_web_start
from cmdbox.app.options import Options
from iinfer.app import web
from pathlib import Path
from typing import Dict, Any, Tuple, List
from urllib.request import pathname2url
import argparse
import logging


class WebStart(cmdbox_web_start.WebStart):
    def get_option(self):
        """
        この機能のオプションを返します

        Returns:
            Dict[str, Any]: オプション
        """
        opt = super().get_option()
        opt["choice"] += [
            dict(opt="showimg_html", type=Options.T_FILE, default=None, required=False, multi=False, hide=False, choice=None, fileio="in",
                 description_ja="`showimg.html` を指定します。省略時はiinfer内蔵のHTMLファイルを使用します。",
                 description_en="Specify `showimg.html`. If omitted, the iinfer built-in HTML file is used."),
            dict(opt="webcap_html", type=Options.T_FILE, default=None, required=False, multi=False, hide=False, choice=None, fileio="in",
                 description_ja="`webcap.html` を指定します。省略時はiinfer内蔵のHTMLファイルを使用します。",
                 description_en="Specify `webcap.html`. If omitted, the iinfer built-in HTML file is used."),
            dict(opt="anno_html", type=Options.T_FILE, default=None, required=False, multi=False, hide=False, choice=None, fileio="in",
                 description_ja="`annotation.html` を指定します。省略時はiinfer内蔵のHTMLファイルを使用します。",
                 description_en="Specify `annotation.html`. If omitted, the iinfer built-in HTML file is used."),
        ]
        return opt

    def createWeb(self, logger:logging.Logger, args:argparse.Namespace) -> web.Web:
        """
        Webオブジェクトを作成します

        Args:
            logger (logging.Logger): ロガー
            args (argparse.Namespace): 引数

        Returns:
            web.Web: Webオブジェクト
        """
        w = web.Web.getInstance(logger, Path(args.data), appcls=self.appcls, ver=self.ver,
                    redis_host=args.host, redis_port=args.port, redis_password=args.password, svname=args.svname,
                    client_only=args.client_only, doc_root=args.doc_root, gui_html=args.gui_html, filer_html=args.filer_html,
                    result_html=args.result_html, users_html=args.users_html,
                    showimg_html=args.showimg_html, webcap_html=args.webcap_html, anno_html=args.anno_html,
                    assets=args.assets, signin_html=args.signin_html, signin_file=args.signin_file, gui_mode=args.gui_mode)
        return w

    def start(self, w:web.Web, logger:logging.Logger, args:argparse.Namespace) -> None:
        """
        Webモードを起動します

        Args:
            w (web.Web): Webオブジェクト
            logger (logging.Logger): ロガー
            args (argparse.Namespace): 引数
        """
        w.start(allow_host=args.allow_host, listen_port=args.listen_port, ssl_listen_port=args.ssl_listen_port,
                ssl_cert=args.ssl_cert, ssl_key=args.ssl_key, ssl_keypass=args.ssl_keypass, ssl_ca_certs=args.ssl_ca_certs,
                session_domain=args.session_domain, session_path=args.session_path,
                session_secure=args.session_secure, session_timeout=args.session_timeout,
                outputs_key=args.outputs_key, gunicorn_workers=args.gunicorn_workers, gunicorn_timeout=args.gunicorn_timeout)


from iinfer.app.injections import after_det_judge_injection
from PIL import Image
from typing import Tuple, Dict, Any


class AfterClsJadgeInjection(after_det_judge_injection.AfterDetJadgeInjection):

    def action(self, reskey:str, name:str, outputs:Dict[str, Any], output_image:Image.Image, session:Dict[str, Any]) -> Tuple[Dict[str, Any], Image.Image]:
        """
        このメソッドは推論を実行した後の処理を実行します。
        Args:
            reskey (str): レスポンスキー
            name (str): モデル名
            outputs (Dict[str, Any]): 推論結果。次の項目が含まれます。
                ・success or warn: 推論成功か警告のキーに対して、その内容が格納されます。
                ・output_image: 推論後の画像データをbase64エンコードした文字列
                ・output_image_shape: 推論後の画像データの形状（base46でコードするときに必要）
                ・output_image_name: クライアントから指定されてきた推論後の画像データの名前

            output_image (Image.Image): 推論後の画像データ
            session (Dict[str, Any]): 推論セッション。次の項目が含まれます。
                ・session: app.predict.Predict#create_session() で作成されたセッション
                ・model_img_width: モデルの入力画像の幅
                ・model_img_height: モデルの入力画像の高さ
                ・predict_obj: app.predict.Predict インスタンス
                ・labels: クラスラベルのリスト
                ・colors: ボックスの色のリスト
                ・tracker: use_trackがTrueの場合、トラッカーのインスタンス
        Returns:
            Tuple[Dict[str, Any], Image.Image]: 後処理後の推論結果と画像データのタプル
        """
        return super().action(reskey, name, outputs, output_image, session)



_base_ = [
    'configs/_base_/schedules/schedule_1x.py', 'configs/_base_/default_runtime.py',
    './yolox_tta.py'
]

img_scale = (640, 640)  # width, height

# model settings
model = dict(
    type='YOLOX',
    data_preprocessor=dict(
        type='DetDataPreprocessor',
        pad_size_divisor=32,
        batch_augments=[
            dict(
                type='BatchSyncRandomResize',
                random_size_range=(480, 800),
                size_divisor=32,
                interval=10)
        ]),
    backbone=dict(
        type='CSPDarknet',
        deepen_factor=0.33,
        widen_factor=0.5,
        out_indices=(2, 3, 4),
        use_depthwise=False,
        spp_kernal_sizes=(5, 9, 13),
        norm_cfg=dict(type='BN', momentum=0.03, eps=0.001),
        act_cfg=dict(type='Swish'),
    ),
    neck=dict(
        type='YOLOXPAFPN',
        in_channels=[128, 256, 512],
        out_channels=128,
        num_csp_blocks=1,
        use_depthwise=False,
        upsample_cfg=dict(scale_factor=2, mode='nearest'),
        norm_cfg=dict(type='BN', momentum=0.03, eps=0.001),
        act_cfg=dict(type='Swish')),
    bbox_head=dict(
        type='YOLOXHead',
        num_classes=80,
        in_channels=128,
        feat_channels=128,
        stacked_convs=2,
        strides=(8, 16, 32),
        use_depthwise=False,
        norm_cfg=dict(type='BN', momentum=0.03, eps=0.001),
        act_cfg=dict(type='Swish'),
        loss_cls=dict(
            type='CrossEntropyLoss',
            use_sigmoid=True,
            reduction='sum',
            loss_weight=1.0),
        loss_bbox=dict(
            type='IoULoss',
            mode='square',
            eps=1e-16,
            reduction='sum',
            loss_weight=5.0),
        loss_obj=dict(
            type='CrossEntropyLoss',
            use_sigmoid=True,
            reduction='sum',
            loss_weight=1.0),
        loss_l1=dict(type='L1Loss', reduction='sum', loss_weight=1.0)),
    train_cfg=dict(assigner=dict(type='SimOTAAssigner', center_radius=2.5)),
    # In order to align the source code, the threshold of the val phase is
    # 0.01, and the threshold of the test phase is 0.001.
    test_cfg=dict(score_thr=0.01, nms=dict(type='nms', iou_threshold=0.65)))

# dataset settings
data_root = 'input/'
dataset_type = 'CocoDataset'

# Example to use different file client
# Method 1: simply set the data root and let the file I/O module
# automatically infer from prefix (not support LMDB and Memcache yet)

# data_root = 's3://openmmlab/datasets/detection/coco/'

# Method 2: Use `backend_args`, `file_client_args` in versions before 3.0.0rc6
# backend_args = dict(
#     backend='petrel',
#     path_mapping=dict({
#         './data/': 's3://openmmlab/datasets/detection/',
#         'data/': 's3://openmmlab/datasets/detection/'
#     }))
backend_args = None

train_pipeline = [
    dict(type='Mosaic', img_scale=img_scale, pad_val=114.0),
    dict(
        type='RandomAffine',
        scaling_ratio_range=(0.1, 2),
        # img_scale is (width, height)
        border=(-img_scale[0] // 2, -img_scale[1] // 2)),
    dict(
        type='MixUp',
        img_scale=img_scale,
        ratio_range=(0.8, 1.6),
        pad_val=114.0),
    dict(type='YOLOXHSVRandomAug'),
    dict(type='RandomFlip', prob=0.5),
    # According to the official implementation, multi-scale
    # training is not considered here but in the
    # 'mmdet/models/detectors/yolox.py'.
    # Resize and Pad are for the last 15 epochs when Mosaic,
    # RandomAffine, and MixUp are closed by YOLOXModeSwitchHook.
    dict(type='Resize', scale=img_scale, keep_ratio=True),
    dict(
        type='Pad',
        pad_to_square=True,
        # If the image is three-channel, the pad value needs
        # to be set separately for each channel.
        pad_val=dict(img=(114.0, 114.0, 114.0))),
    dict(type='FilterAnnotations', min_gt_bbox_wh=(1, 1), keep_empty=False),
    dict(type='PackDetInputs')
]

train_dataset = dict(
    # use MultiImageMixDataset wrapper to support mosaic and mixup
    type='MultiImageMixDataset',
    dataset=dict(
        type=dataset_type,
        data_root=data_root,
        ann_file='train.json',
        data_prefix=dict(img='train/'),
        pipeline=[
            dict(type='LoadImageFromFile', backend_args=backend_args),
            dict(type='LoadAnnotations', with_bbox=True)
        ],
        filter_cfg=dict(filter_empty_gt=False, min_size=32),
        backend_args=backend_args),
    pipeline=train_pipeline)

test_pipeline = [
    dict(type='LoadImageFromFile', backend_args=backend_args),
    dict(type='Resize', scale=img_scale, keep_ratio=True),
    dict(
        type='Pad',
        pad_to_square=True,
        pad_val=dict(img=(114.0, 114.0, 114.0))),
    dict(type='LoadAnnotations', with_bbox=True),
    dict(
        type='PackDetInputs',
        meta_keys=('img_id', 'img_path', 'ori_shape', 'img_shape',
                   'scale_factor'))
]

train_dataloader = dict(
    batch_size=8,
    num_workers=4,
    persistent_workers=True,
    sampler=dict(type='DefaultSampler', shuffle=True),
    dataset=train_dataset)
val_dataloader = dict(
    batch_size=8,
    num_workers=4,
    persistent_workers=True,
    drop_last=False,
    sampler=dict(type='DefaultSampler', shuffle=False),
    dataset=dict(
        type=dataset_type,
        data_root=data_root,
        ann_file='val.json',
        data_prefix=dict(img='val/'),
        test_mode=True,
        pipeline=test_pipeline,
        backend_args=backend_args))
test_dataloader = val_dataloader

val_evaluator = dict(
    type='CocoMetric',
    ann_file=data_root + 'val.json',
    metric='bbox',
    backend_args=backend_args)
test_evaluator = val_evaluator

# training settings
max_epochs = 300
num_last_epochs = 15
interval = 10

train_cfg = dict(max_epochs=max_epochs, val_interval=interval)

# optimizer
# default 8 gpu
base_lr = 0.01
optim_wrapper = dict(
    type='OptimWrapper',
    optimizer=dict(
        type='SGD', lr=base_lr, momentum=0.9, weight_decay=5e-4,
        nesterov=True),
    paramwise_cfg=dict(norm_decay_mult=0., bias_decay_mult=0.))

# learning rate
param_scheduler = [
    dict(
        # use quadratic formula to warm up 5 epochs
        # and lr is updated by iteration
        # TODO: fix default scope in get function
        type='mmdet.QuadraticWarmupLR',
        by_epoch=True,
        begin=0,
        end=5,
        convert_to_iter_based=True),
    dict(
        # use cosine lr from 5 to 285 epoch
        type='CosineAnnealingLR',
        eta_min=base_lr * 0.05,
        begin=5,
        T_max=max_epochs - num_last_epochs,
        end=max_epochs - num_last_epochs,
        by_epoch=True,
        convert_to_iter_based=True),
    dict(
        # use fixed lr during last 15 epochs
        type='ConstantLR',
        by_epoch=True,
        factor=1,
        begin=max_epochs - num_last_epochs,
        end=max_epochs,
    )
]

default_hooks = dict(
    checkpoint=dict(
        interval=interval,
        max_keep_ckpts=3  # only keep latest 3 checkpoints
    ))

custom_hooks = [
    dict(
        type='YOLOXModeSwitchHook',
        num_last_epochs=num_last_epochs,
        priority=48),
    dict(type='SyncNormHook', priority=48),
    dict(
        type='EMAHook',
        ema_type='ExpMomentumEMA',
        momentum=0.0001,
        update_buffers=True,
        priority=49)
]

# NOTE: `auto_scale_lr` is for automatically scaling LR,
# USER SHOULD NOT CHANGE ITS VALUES.
# base_batch_size = (8 GPUs) x (8 samples per GPU)
auto_scale_lr = dict(base_batch_size=64)



from cmdbox.app.features.web import cmdbox_web_gui
from iinfer import version


class Gui(cmdbox_web_gui.Gui):
    def __init__(self, appcls, ver):
        super().__init__(appcls, ver)
        self.version_info.append(dict(tabid='versions_iinfer', title=version.__appid__,
                                      thisapp=True if version.__appid__ == ver.__appid__ else False,
                                      icon=f'assets/iinfer/icon.png', url='versions_iinfer'))



_base_ = './yolox_s_8xb8-300e_coco.py'

# model settings
model = dict(
    backbone=dict(deepen_factor=1.0, widen_factor=1.0),
    neck=dict(
        in_channels=[256, 512, 1024], out_channels=256, num_csp_blocks=3),
    bbox_head=dict(in_channels=256, feat_channels=256))



from cmdbox.app.commons import convert
from iinfer.app import injection
from PIL import Image
from typing import Tuple, Dict, Any
import datetime
import io
import requests
import time

class AfterHttpInjection(injection.AfterInjection):

    def action(self, reskey:str, name:str, outputs:Dict[str, Any], output_image:Image.Image, session:Dict[str, Any]) -> Tuple[Dict[str, Any], Image.Image]:
        """
        このメソッドは推論を実行した後の処理を実行します。
        Args:
            reskey (str): レスポンスキー
            name (str): モデル名
            outputs (Dict[str, Any]): 推論結果。次の項目が含まれます。
                ・success or warn: 推論成功か警告のキーに対して、その内容が格納されます。
                ・output_image: 推論後の画像データをbase64エンコードした文字列
                ・output_image_shape: 推論後の画像データの形状（base46でコードするときに必要）
                ・output_image_name: クライアントから指定されてきた推論後の画像データの名前

            output_image (Image.Image): 推論後の画像データ
            session (Dict[str, Any]): 推論セッション。次の項目が含まれます。
                ・session: app.predict.Predict#create_session() で作成されたセッション
                ・model_img_width: モデルの入力画像の幅
                ・model_img_height: モデルの入力画像の高さ
                ・predict_obj: app.predict.Predict インスタンス
                ・labels: クラスラベルのリスト
                ・colors: ボックスの色のリスト
                ・tracker: use_trackがTrueの場合、トラッカーのインスタンス
        Returns:
            Tuple[Dict[str, Any], Image.Image]: 後処理後の推論結果と画像データのタプル
        """
        #import http
        #http.client.HTTPConnection.debuglevel=1
        self.req_session = requests.Session()
        try:
            url = self.get_config('outputs_url', None)
            if url is not None:
                json_without_img = self.get_config('json_without_img', False)
                if json_without_img:
                    del outputs['output_image']
                    del outputs['output_image_shape']
                result = self.post_json(url, outputs)
                self.add_success(outputs, result)
        except Exception as e:
            self.add_warning(outputs, str(e))

        try:
            url = self.get_config('output_image_url', None)
            if url is not None:
                result = self.post_img(url, outputs, output_image)
                self.add_success(outputs, result)
        except Exception as e:
            self.add_warning(outputs, str(e))

        return outputs, output_image

    def post_json(self, url, outputs:Dict[str, Any]):
        tm = time.time()
        res = self.req_session.post(url, json=outputs, verify=False, timeout=30)
        if res.status_code != 200:
            raise Exception(f"HTTP POST request failed. status_code={res.status_code} res.reason={res.reason} res.text={res.text} url={url}")
        try:
            result = res.json()
        except:
            result = dict(success=res.text)
        return result

    def post_img(self, url, outputs:Dict[str, Any], output_image:Image.Image):
        finename = self.get_config('fileup_name', None)
        ext = self.get_config('output_image_ext', 'jpeg')
        if finename is None:
            prefix = self.get_config('output_image_prefix', 'output_')
            finename = f'{prefix}{datetime.datetime.now().strftime("%Y%m%d%H%M%S")}.{ext}'

        finename = finename if 'output_image_name' not in outputs else outputs['output_image_name']
        finename = finename.strip() if finename is not None else finename
        img_bytes = convert.img2byte(output_image, format=ext)
        file = {'file': (finename, io.BytesIO(img_bytes))}

        res = self.req_session.post(url, files=file, verify=False, timeout=30)
        if res.status_code != 200:
            raise Exception(f"HTTP POST request failed. status_code={res.status_code} res.reason={res.reason} res.text={res.text} url={url}")
        try:
            result = res.json()
        except:
            result = dict(success=res.text)
        return result



from cmdbox.app.commons import convert
from cmdbox.app.features.web import cmdbox_web_exec_cmd
from iinfer.app.web import Web
from fastapi import FastAPI, Request, Response, HTTPException
from pathlib import Path
import tempfile
import traceback
import zipfile


class AnnoArchive(cmdbox_web_exec_cmd.ExecCmd):
    def route(self, web:Web, app:FastAPI) -> None:
        """
        webモードのルーティングを設定します

        Args:
            web (Web): Webオブジェクト
            app (FastAPI): FastAPIオブジェクト
        """
        @app.post('/annotation/archive/{constr}')
        async def anno_archive(req:Request, res:Response, constr:str):
            signin = web.signin.check_signin(req, res)
            if signin is not None:
                raise HTTPException(status_code=401, detail=self.DEFAULT_401_MESSAGE)
            try:
                host, port, svname, password, output_path, scope, img_thumbnail = convert.b64str2str(constr).split('\t')
                output_path = output_path[1:] if output_path.startswith('/') else output_path
                form = await req.form()
                archives = await req.json()
                data_dir = web.data if scope == 'client' else Path.cwd()
                opt = dict(host=host, port=port, svname=svname, password=password, scope=scope,
                            img_thumbnail=img_thumbnail, mode='client', cmd='file_download', client_data=data_dir, stdout_log=False)
                opt['capture_stdout'] = nothread = True
                with tempfile.TemporaryDirectory() as tmpdir:
                    tmpdir_path = Path(tmpdir)
                    for arc in archives:
                        opt['svpath'] = arc
                        ret = self.exec_cmd(req, res, web, 'file_download', opt, nothread, self.appcls)
                        if len(ret) == 0 or 'success' not in ret[0]:
                            return ret
                        arc = arc[1:] if arc.startswith('/') else arc
                        p:Path = Path(tmpdir_path / arc)
                        if not p.parent.exists():
                            p.parent.mkdir(parents=True)
                        with open(p, 'wb') as f:
                            f.write(convert.b64str2bytes(ret[0]['success']['data']))

                    upload_file:Path = tmpdir_path / output_path
                    if not str(upload_file).startswith(str(tmpdir_path)):
                        web.logger.warning(f'Invalid output path. {output_path}')
                        return dict(warn=f'Invalid output path. {output_path}')
                    with zipfile.ZipFile(upload_file, 'w') as zipf:
                        dir = str(Path(output_path).parent).replace('\\', '/') + '/'
                        for arc in archives:
                            arc = arc[1:] if arc.startswith('/') else arc
                            zipf.write(tmpdir_path / arc, arc.replace(dir, ''))

                    opt['cmd'] = 'file_upload'
                    opt['svpath'] = output_path
                    opt['orverwrite'] = True
                    opt['upload_file'] = str(upload_file).replace('"','')
                    web.options.audit_exec(req, res, web)
                    ret = self.exec_cmd(req, res, web, "file_upload", opt, nothread=True, appcls=self.appcls)
                    if len(ret) == 0 or 'success' not in ret[0]:
                        return ret
                    return ret
            except Exception as e:
                web.logger.warning(f'anno_archive error {e}')
                return dict(warn=f'anno_archive error. {traceback.format_exc()}')



_base_ = [
    'configs/_base_/models/san_vit-b16.py', 'configs/_base_/datasets/coco-stuff164k.py',
    'configs/_base_/default_runtime.py', 'configs/_base_/schedules/schedule_160k.py'
]
crop_size = (640, 640)
train_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(type='LoadAnnotations'),
    dict(
        type='RandomChoiceResize',
        scales=[int(640 * x * 0.1) for x in range(5, 16)],
        resize_type='ResizeShortestEdge',
        max_size=2560),
    dict(type='RandomCrop', crop_size=crop_size, cat_max_ratio=1.0),
    dict(type='PhotoMetricDistortion'),
    dict(type='RandomFlip', prob=0.5),
    dict(type='PackSegInputs')
]

test_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(type='ResizeShortestEdge', scale=crop_size, max_size=2560),
    dict(type='LoadAnnotations'),
    dict(type='PackSegInputs')
]

# By default, models are trained on 4 GPUs with 8 images per GPU
train_dataloader = dict(batch_size=8, dataset=dict(pipeline=train_pipeline))
val_dataloader = dict(batch_size=1, dataset=dict(pipeline=test_pipeline))
test_dataloader = val_dataloader

pretrained = 'https://download.openmmlab.com/mmsegmentation/v0.5/san/clip_vit-base-patch16-224_3rdparty-d08f8887.pth'  # noqa
data_preprocessor = dict(
    mean=[122.7709, 116.7460, 104.0937],
    std=[68.5005, 66.6322, 70.3232],
    size_divisor=640,
    test_cfg=dict(size_divisor=32))
model = dict(
    pretrained=pretrained,
    text_encoder=dict(dataset_name='coco-stuff164k'),
    decode_head=dict(num_classes=171))

# training schedule for 60k
train_cfg = dict(
    type='IterBasedTrainLoop',
    max_iters=60000,
    val_interval=500,
    val_begin=55000)
default_hooks = dict(
    checkpoint=dict(
        type='CheckpointHook',
        by_epoch=False,
        interval=10000,
        save_best='mIoU'))

# AdamW optimizer, no weight decay for position embedding & layer norm
# in backbone
optim_wrapper = dict(
    _delete_=True,
    type='AmpOptimWrapper',
    optimizer=dict(
        type='AdamW', lr=0.0001, betas=(0.9, 0.999), weight_decay=0.0001),
    paramwise_cfg=dict(
        custom_keys={
            'img_encoder': dict(lr_mult=0.1, decay_mult=1.0),
            'pos_embed': dict(decay_mult=0.),
            'cls_token': dict(decay_mult=0.),
            'norm': dict(decay_mult=0.)
        }),
    loss_scale='dynamic',
    clip_grad=dict(max_norm=0.01, norm_type=2))

param_scheduler = [
    dict(
        type='PolyLR',
        eta_min=0.0,
        power=1.0,
        begin=0,
        end=60000,
        by_epoch=False,
    )
]



from cmdbox.app import common, feature
from cmdbox.app.commons import convert, redis_client
from cmdbox.app.options import Options
from motpy import Detection
from iinfer.app import common as cmn, client, injection, predict
from pathlib import Path
from typing import Dict, Any, Tuple, List, Union
from PIL import Image
import argparse
import cv2
import logging
import sys
import time


class ClientPredict(feature.Feature):
    def get_mode(self) -> Union[str, List[str]]:
        """
        この機能のモードを返します

        Returns:
            Union[str, List[str]]: モード
        """
        return 'client'

    def get_cmd(self):
        """
        この機能のコマンドを返します

        Returns:
            str: コマンド
        """
        return 'predict'
    
    def get_option(self):
        """
        この機能のオプションを返します

        Returns:
            Dict[str, Any]: オプション
        """
        return dict(
            use_redis=self.USE_REDIS_TRUE, nouse_webmode=False,
            description_ja="AIモデルを指定して推論を実行します。",
            description_en="Perform inference by specifying the AI model.",
            choice=[
                dict(opt="host", type=Options.T_STR, default=self.default_host, required=True, multi=False, hide=True, choice=None, web="mask",
                     description_ja="Redisサーバーのサービスホストを指定します。",
                     description_en="Specify the service host of the Redis server."),
                dict(opt="port", type=Options.T_INT, default=self.default_port, required=True, multi=False, hide=True, choice=None, web="mask",
                     description_ja="Redisサーバーのサービスポートを指定します。",
                     description_en="Specify the service port of the Redis server."),
                dict(opt="password", type=Options.T_STR, default=self.default_pass, required=True, multi=False, hide=True, choice=None, web="mask",
                     description_ja="Redisサーバーのアクセスパスワード(任意)を指定します。省略時は `password` を使用します。",
                     description_en="Specify the access password of the Redis server (optional). If omitted, `password` is used."),
                dict(opt="svname", type=Options.T_STR, default=self.default_svname, required=True, multi=False, hide=True, choice=None, web="readonly",
                     description_ja="サーバーのサービス名を指定します。省略時は `server` を使用します。",
                     description_en="Specify the service name of the inference server. If omitted, `server` is used."),
                dict(short="n", opt="name", type=Options.T_STR, default=None, required=True, multi=False, hide=False, choice=None,
                     description_ja="削除するAIモデルの登録名を指定します。",
                     description_en="Specify the registration name of the AI model to be deleted.",
                     test_true={"yolox":"yolox",
                                "upernet":"upernet",
                                "san":"san",
                                "pspnet":"pspnet",
                                "swin":"swin",
                                "lnsightface":"lnsightface",
                                "yolo3":"yolo3",
                                "effnet":"effnet",
                                "custom":"custom"}),
                dict(short="i", opt="input_file", type=Options.T_FILE, default=None, required=False, multi=False, hide=False, choice=None, fileio="in",
                     description_ja="推論させる画像をファイルで指定します。",
                     description_en="Specify the image to be inferred by file.",
                     test_true={"yolox":"tests/dog.jpg",
                                "upernet":None,
                                "san":"tests/dog.jpg",
                                "pspnet":"tests/dog.jpg",
                                "swin":"tests/dog.jpg",
                                "lnsightface":"capture.csv",
                                "yolo3":"tests/dog.jpg",
                                "effnet":"tests/dog.jpg",
                                "custom":None}),
                dict(opt="stdin", type=Options.T_BOOL, default=False, required=False, multi=False, hide=False, choice=[True, False],
                     description_ja="推論させる画像を標準入力から読み込む。",
                     description_en="Read the image to be inferred from standard input.",
                     test_true={"yolox":False,
                                "upernet":True,
                                "custom":True},
                     test_stdin={"yolox":None,
                                 "upernet":"tests/dog.jpg",
                                 "custom":"tests/dog.jpg"}),
                dict(opt="nodraw", type=Options.T_BOOL, default=False, required=False, multi=False, hide=False, choice=[True, False],
                     description_ja="推論結果画像にbbox等の描き込みを行わない。",
                     description_en="Do not draw bboxes etc. on the inference result image.",
                     test_true={"yolox":True}),
                dict(opt="pred_input_type", type=Options.T_STR, default="jpeg", required=True, multi=False, hide=False,
                        choice=['bmp', 'png', 'jpeg', 'capture', 'output_json'],
                     description_ja="推論させる入力タイプを指定します。",
                     description_en="Specifies the input type to be inferred.",
                     test_true={"yolox":"jpeg",
                                "upernet":"jpeg",
                                "san":"jpeg",
                                "pspnet":"jpeg",
                                "swin":"jpeg",
                                "lnsightface":"capture",
                                "yolo3":"jpeg",
                                "effnet":"jpeg",
                                "custom":"jpeg"}),
                dict(opt="output_image", type=Options.T_FILE, default=None, required=False, multi=False, hide=False, choice=None, fileio="out",
                     description_ja="推論結果画像の保存先ファイルを指定します。",
                     description_en="Specify the destination file for saving the inference result image.",
                     test_true={"yolox":"pred.jpg"}),
                dict(short="P", opt="output_preview", type=Options.T_BOOL, default=False, required=False, multi=False, hide=False, choice=[True, False],
                     description_ja="推論結果画像を `cv2.imshow` で表示します。",
                     description_en="Display the inference result image with `cv2.imshow`.",
                     test_true={"yolox":True,
                                "upernet":True,
                                "san":True,
                                "pspnet":True,
                                "swin":True,
                                "lnsightface":True,
                                "yolo3":True,
                                "effnet":True,
                                "custom":True}),
                dict(opt="retry_count", type=Options.T_INT, default=3, required=False, multi=False, hide=True, choice=None,
                     description_ja="Redisサーバーへの再接続回数を指定します。0以下を指定すると永遠に再接続を行います。",
                     description_en="Specifies the number of reconnections to the Redis server.If less than 0 is specified, reconnection is forever."),
                dict(opt="retry_interval", type=Options.T_INT, default=5, required=False, multi=False, hide=True, choice=None,
                     description_ja="Redisサーバーに再接続までの秒数を指定します。",
                     description_en="Specifies the number of seconds before reconnecting to the Redis server."),
                dict(opt="timeout", type=Options.T_INT, default="15", required=False, multi=False, hide=True, choice=None,
                     description_ja="サーバーの応答が返ってくるまでの最大待ち時間を指定。",
                     description_en="Specify the maximum waiting time until the server responds.",
                     test_true={"yolox":15,
                                "upernet":60,
                                "san":60,
                                "pspnet":60,
                                "swin":60,
                                "lnsightface":15,
                                "yolo3":15,
                                "effnet":15,
                                "custom":15}),
                dict(opt="output_json", short="o" , type=Options.T_FILE, default=None, required=False, multi=False, hide=True, choice=None, fileio="out",
                     description_ja="処理結果jsonの保存先ファイルを指定。",
                     description_en="Specify the destination file for saving the processing result json."),
                dict(opt="output_json_append", short="a" , type=Options.T_BOOL, default=False, required=False, multi=False, hide=True, choice=[True, False],
                     description_ja="処理結果jsonファイルを追記保存します。",
                     description_en="Save the processing result json file by appending."),
                dict(opt="stdout_log", type=Options.T_BOOL, default=True, required=False, multi=False, hide=True, choice=[True, False],
                     description_ja="GUIモードでのみ使用可能です。コマンド実行時の標準出力をConsole logに出力します。",
                     description_en="Available only in GUI mode. Outputs standard output during command execution to Console log."),
                dict(opt="capture_stdout", type=Options.T_BOOL, default=True, required=False, multi=False, hide=True, choice=[True, False],
                     description_ja="GUIモードでのみ使用可能です。コマンド実行時の標準出力をキャプチャーし、実行結果画面に表示します。",
                     description_en="Available only in GUI mode. Captures standard output during command execution and displays it on the execution result screen."),
                dict(opt="capture_maxsize", type=Options.T_INT, default=self.DEFAULT_CAPTURE_MAXSIZE, required=False, multi=False, hide=True, choice=None,
                     description_ja="GUIモードでのみ使用可能です。コマンド実行時の標準出力の最大キャプチャーサイズを指定します。",
                     description_en="Available only in GUI mode. Specifies the maximum capture size of standard output when executing commands."),
            ]
        )
    
    def get_svcmd(self):
        """
        この機能のサーバー側のコマンドを返します

        Returns:
            str: サーバー側のコマンド
        """
        return 'predict'

    def apprun(self, logger:logging.Logger, args:argparse.Namespace, tm:float, pf:List[Dict[str, float]]=[]) -> Tuple[int, Dict[str, Any], Any]:
        """
        この機能の実行を行います

        Args:
            logger (logging.Logger): ロガー
            args (argparse.Namespace): 引数
            tm (float): 実行開始時間
            pf (List[Dict[str, float]]): 呼出元のパフォーマンス情報

        Returns:
            Tuple[int, Dict[str, Any], Any]: 終了コード, 結果, オブジェクト
        """
        if args.svname is None:
            msg = {"warn":f"Please specify the --svname option."}
            common.print_format(msg, args.format, tm, args.output_json, args.output_json_append, pf=pf)
            return self.RESP_WARN, msg, None
        cl = client.Client(logger, redis_host=args.host, redis_port=args.port, redis_password=args.password, svname=args.svname)

        try:
            if args.input_file is not None:
                if logger.level == logging.DEBUG:
                    logger.debug(f"app.main: args.mode={args.mode}, args.cmd={args.cmd}, args.name={args.name}, args.input_file={args.input_file}")
                ret = cl.predict(args.name, image_file=args.input_file, pred_input_type=args.pred_input_type,
                                        output_image_file=args.output_image, output_preview=args.output_preview, nodraw=args.nodraw,
                                        retry_count=args.retry_count, retry_interval=args.retry_interval, timeout=args.timeout)
                if type(ret) is list:
                    for r in ret:
                        common.print_format(r, args.format, tm, args.output_json, args.output_json_append, pf=pf)
                        if logger.level == logging.DEBUG:
                            ret_str = common.to_str(r, slise=100)
                            logger.debug(f"app.main: args.mode={args.mode}, args.cmd={args.cmd}, ret={ret_str}")
                        tm = time.perf_counter()
                        args.output_json_append = True
                else:
                    common.print_format(ret, args.format, tm, args.output_json, args.output_json_append, pf=pf)
            elif args.stdin:
                if args.pred_input_type is None:
                    msg = {"warn":f"Please specify the --pred_input_type option."}
                    common.print_format(msg, args.format, tm, args.output_json, args.output_json_append, pf=pf)
                    if logger.level == logging.DEBUG:
                        msg_str = common.to_str(msg, slise=100)
                        logger.debug(f"app.main: args.mode={args.mode}, args.cmd={args.cmd}, msg={msg_str}")
                    return self.RESP_WARN, msg, cl
                if args.pred_input_type in ['capture']:
                    def _pred(args, line, tm):
                        if logger.level == logging.DEBUG:
                            line_str = common.to_str(line, slise=100)
                            logger.debug(f"app.main: args.mode={args.mode}, args.cmd={args.cmd}, args.name={args.name}, image={line_str}")
                        ret = cl.predict(args.name, image=line, pred_input_type=args.pred_input_type,
                                            output_image_file=args.output_image, output_preview=args.output_preview, nodraw=args.nodraw,
                                            retry_count=args.retry_count, retry_interval=args.retry_interval, timeout=args.timeout)
                        common.print_format(ret, args.format, tm, args.output_json, args.output_json_append, pf=pf)
                        if logger.level == logging.DEBUG:
                            ret_str = common.to_str(ret, slise=100)
                            logger.debug(f"app.main: args.mode={args.mode}, args.cmd={args.cmd}, ret={ret_str}")
                    for line in sys.stdin:
                        # 標準入力による推論処理は非同期で行う(同名複数serverの場合にスループットを向上させるため)
                        #thread = threading.Thread(target=_pred, args=(args, line, tm))
                        #thread.start()
                        _pred(args, line, tm)
                        tm = time.perf_counter()
                        args.output_json_append = True
                else:
                    if logger.level == logging.DEBUG:
                        logger.debug(f"app.main: args.mode={args.mode}, args.cmd={args.cmd}, args.name={args.name}, image=<stdin>")
                    ret = cl.predict(args.name, image=sys.stdin.buffer.read(), pred_input_type=args.pred_input_type,
                                        output_image_file=args.output_image, output_preview=args.output_preview, nodraw=args.nodraw,
                                        retry_count=args.retry_count, retry_interval=args.retry_interval, timeout=args.timeout)
                    common.print_format(ret, args.format, tm, args.output_json, args.output_json_append, pf=pf)
                    if logger.level == logging.DEBUG:
                        ret_str = common.to_str(ret, slise=100)
                        logger.debug(f"app.main: args.mode={args.mode}, args.cmd={args.cmd}, ret={ret_str}")
                    tm = time.perf_counter()
            else:
                msg = {"warn":f"Image file or stdin is empty."}
                common.print_format(msg, args.format, tm, args.output_json, args.output_json_append, pf=pf)
                if logger.level == logging.DEBUG:
                    msg_str = common.to_str(msg, slise=100)
                    logger.debug(f"app.main: args.mode={args.mode}, args.cmd={args.cmd}, msg={msg_str}")
                return self.RESP_WARN, msg, cl
        finally:
            try:
                cv2.destroyWindow('preview')
            except:
                pass

        return self.RESP_SUCCESS, None, cl

    def is_cluster_redirect(self):
        """
        クラスター宛のメッセージの場合、メッセージを転送するかどうかを返します

        Returns:
            bool: メッセージを転送する場合はTrue
        """
        return False

    def svrun(self, data_dir:Path, logger:logging.Logger, redis_cli:redis_client.RedisClient, msg:List[str],
              sessions:Dict[str, Dict[str, Any]]) -> int:
        """
        この機能のサーバー側の実行を行います

        Args:
            data_dir (Path): データディレクトリ
            logger (logging.Logger): ロガー
            redis_cli (redis_client.RedisClient): Redisクライアント
            msg (List[str]): 受信メッセージ
            sessions (Dict[str, Dict[str, Any]]): セッション情報
        
        Returns:
            int: 終了コード
        """
        nodraw = True if msg[4] == 'True' else False
        shape = [int(msg[5]), int(msg[6])]
        if int(msg[7]) > 0: shape.append(int(msg[7]))
        output_image_name = msg[8]
        if shape[0] >0 and shape[1] > 0:
            img_npy = convert.b64str2npy(msg[3], shape)
            image = convert.npy2img(img_npy)
            st = self.predict(msg[1], msg[2], image, output_image_name, nodraw,
                              data_dir, logger, redis_cli, sessions)
        else:
            st = self.predict(msg[1], msg[2], convert.b64str2str(msg[3]), output_image_name, nodraw,
                              data_dir, logger, redis_cli, sessions)

        return st

    def predict(self, reskey:str, name:str, input_data:Union[Image.Image, str], output_image_name:str, nodraw:bool,
                data_dir:Path, logger:logging.Logger, redis_cli:redis_client.RedisClient, sessions:Dict[str, Dict[str, Any]]) -> int:
        """
        クライアントから送られてきた画像の推論を行う。

        Args:
            reskey (str): レスポンスキー
            name (dict): モデル名
            input_data (Image.Image | str): 推論するデータ
            output_image_name (str): 出力画像のファイル名
            nodraw (bool): 描画フラグ
            data_dir (Path): データディレクトリ
            logger (logging.Logger): ロガー
            redis_cli (redis_client.RedisClient): Redisクライアント
            sessions (Dict[str, Dict[str, Any]]): セッション情報
        """
        if name is None or name == "":
            logger.warning(f"Name is empty.")
            redis_cli.rpush(reskey, {"warn": f"Name is empty."})
            return self.RESP_WARN
        if input_data is None:
            logger.warning(f"input_data is empty.")
            redis_cli.rpush(reskey, {"warn": f"input_data is empty."})
            return self.RESP_WARN
        if name not in sessions:
            logger.warning(f"{name} has not yet started a session.")
            redis_cli.rpush(reskey, {"warn": f"{name} has not yet started a session."})
            return self.RESP_WARN
        if nodraw is None:
            nodraw = False
        session = sessions[name]
        try:
            predict_process_start = time.perf_counter()
            # 前処理を実行
            if session['before_injections'] is not None:
                injections:List[injection.BeforeInjection] = session['before_injections']
                for inject in injections:
                    input_data = inject.action(reskey, name, input_data, session)
            before_injections_end = time.perf_counter()
            # 推論を実行
            predict_obj:predict.Predict = session['predict_obj']
            outputs, output_image = predict_obj.predict(session['session'], session['model_img_width'], session['model_img_height'], input_data,
                                                        labels=session['labels'], colors=session['colors'], nodraw=nodraw)
            outputs['image_name'] = output_image_name
            predict_end = time.perf_counter()
            if session['tracker'] is not None:
                if 'output_boxes' in outputs and 'output_scores' in outputs and 'output_classes' in outputs:
                    detections = [Detection(box, score, cls) for box, score, cls in zip(outputs['output_boxes'], outputs['output_scores'], outputs['output_classes'])]
                    session['tracker'].step(detections=detections)
                    tracks = session['tracker'].active_tracks()
                    outputs['output_tracks'] = [t.id for t in tracks]
                    if output_image is not None and not nodraw:
                        output_image, _ = cmn.draw_boxes(output_image, outputs['output_boxes'], outputs['output_scores'], outputs['output_classes'], ids=outputs['output_tracks'])
            tracker_end = time.perf_counter()

            def _after_injection(reskey:str, name:str, output:dict, output_image:Image.Image, session:dict):
                if session['after_injections'] is not None:
                    injections:List[injection.AfterInjection] = session['after_injections']
                    for inject in injections:
                        output, output_image = inject.action(reskey, name, output, output_image, session)
                return output, output_image

            def _set_perftime(output, predict_process_start, before_injections_end, predict_end, tracker_end, after_injections_end, predict_process_end):
                performance = [dict(key="sv_before", val=f"{(before_injections_end-predict_process_start):.3f}s"),
                               dict(key="sv_predict", val=f"{(predict_end-before_injections_end):.3f}s"),
                               dict(key="sv_track", val=f"{(tracker_end-predict_end):.3f}s"),
                               dict(key="sv_after", val=f"{(after_injections_end-tracker_end):.3f}s"),
                               dict(key="sv_process", val=f"{(predict_process_end-predict_process_start):.3f}s")]
                if 'success' in output:
                    output['success']['performance'] = performance

            if output_image is not None:
                output = dict(success=outputs, output_image_name=output_image_name)
                output_image_npy = convert.img2npy(output_image)
                output_image_b64 = convert.npy2b64str(output_image_npy)
                output['output_image'] = output_image_b64
                output['output_image_shape'] = output_image_npy.shape
                predict_process_end = time.perf_counter()
                # 後処理を実行
                output, output_image = _after_injection(reskey, name, output, output_image, session)
                after_injections_end = time.perf_counter()
                _set_perftime(output, predict_process_start, before_injections_end, predict_end,
                             tracker_end, after_injections_end, predict_process_end)
                redis_cli.rpush(reskey, output)
                return self.RESP_SUCCESS
            output = dict(success=outputs)
            # 後処理を実行
            output, _ = _after_injection(reskey, name, output, None, session)
            after_injections_end = predict_process_end = time.perf_counter()
            _set_perftime(output, predict_process_start, before_injections_end, predict_end,
                            tracker_end, after_injections_end, predict_process_end)
            redis_cli.rpush(reskey, output)
            return self.RESP_SUCCESS
        except Exception as e:
            logger.warning(f"Failed to run inference: {e}", exc_info=True)
            redis_cli.rpush(reskey, {"warn": f"Failed to run inference: {e}"})
            return self.RESP_WARN



from iinfer.app.predicts.mmseg_seg_PSPNet import MMSegPSPNet
import logging

SITE = 'https://github.com/open-mmlab/mmsegmentation/tree/main/configs/swin'
IMAGE_WIDTH = 512
IMAGE_HEIGHT = 512
REQUIREd_MODEL_CONF = True
REQUIREd_MODEL_WEIGHT = False

class MMSegSwinUpernet(MMSegPSPNet):
    def __init__(self, logger:logging.Logger) -> None:
        super().__init__(logger)



from cmdbox.app import common, feature
from cmdbox.app.options import Options
from iinfer.app import common as cmn
from typing import Dict, Any, Tuple, Union, List
import argparse
import logging


class ClientTrainTypeList(feature.OneshotResultEdgeFeature):
    def get_mode(self) -> Union[str, List[str]]:
        """
        この機能のモードを返します

        Returns:
            Union[str, List[str]]: モード
        """
        return 'client'

    def get_cmd(self):
        """
        この機能のコマンドを返します

        Returns:
            str: コマンド
        """
        return 'train_type_list'
    
    def get_option(self):
        """
        この機能のオプションを返します

        Returns:
            Dict[str, Any]: オプション
        """
        return dict(
            use_redis=self.USE_REDIS_FALSE, nouse_webmode=False,
            description_ja="学習タイプ一覧を取得します。",
            description_en="Get a list of train types.",
            choice=[
                dict(opt="output_json", short="o" , type=Options.T_FILE, default=None, required=False, multi=False, hide=True, choice=None, fileio="out",
                     description_ja="処理結果jsonの保存先ファイルを指定。",
                     description_en="Specify the destination file for saving the processing result json.",
                     test_true={"yolox":None}),
                dict(opt="output_json_append", short="a" , type=Options.T_BOOL, default=False, required=False, multi=False, hide=True, choice=[True, False],
                     description_ja="処理結果jsonファイルを追記保存します。",
                     description_en="Save the processing result json file by appending."),
                dict(opt="stdout_log", type=Options.T_BOOL, default=True, required=False, multi=False, hide=True, choice=[True, False],
                     description_ja="GUIモードでのみ使用可能です。コマンド実行時の標準出力をConsole logに出力します。",
                     description_en="Available only in GUI mode. Outputs standard output during command execution to Console log."),
                dict(opt="capture_stdout", type=Options.T_BOOL, default=True, required=False, multi=False, hide=True, choice=[True, False],
                     description_ja="GUIモードでのみ使用可能です。コマンド実行時の標準出力をキャプチャーし、実行結果画面に表示します。",
                     description_en="Available only in GUI mode. Captures standard output during command execution and displays it on the execution result screen."),
                dict(opt="capture_maxsize", type=Options.T_INT, default=self.DEFAULT_CAPTURE_MAXSIZE, required=False, multi=False, hide=True, choice=None,
                     description_ja="GUIモードでのみ使用可能です。コマンド実行時の標準出力の最大キャプチャーサイズを指定します。",
                     description_en="Available only in GUI mode. Specifies the maximum capture size of standard output when executing commands."),
            ]
        )

    def apprun(self, logger:logging.Logger, args:argparse.Namespace, tm:float, pf:List[Dict[str, float]]=[]) -> Tuple[int, Dict[str, Any], Any]:
        """
        この機能の実行を行います

        Args:
            logger (logging.Logger): ロガー
            args (argparse.Namespace): 引数
            tm (float): 実行開始時間
            pf (List[Dict[str, float]]): 呼出元のパフォーマンス情報

        Returns:
            Tuple[int, Dict[str, Any], Any]: 終了コード, 結果, オブジェクト
        """
        type_list = [dict(train_type=key, site=val['site'],
                            model_type=f"{val['model_type'].__module__}.{val['model_type'].__name__}") for key,val in cmn.BASE_TRAIN_MODELS.items()]
        type_list.append(dict(train_type='Custom', site='Custom'))
        ret = dict(success=type_list)
        common.print_format(ret, args.format, tm, args.output_json, args.output_json_append, pf=pf)
        return self.RESP_SUCCESS, ret, None



from iinfer.app import postprocess
from iinfer.app.injections import after_showimg_injection
from PIL import Image
from typing import Dict, Tuple, Any
from urllib3.exceptions import InsecureRequestWarning
import logging
import urllib3
urllib3.disable_warnings(InsecureRequestWarning)


class Showimg(postprocess.Postprocess):
    def __init__(self, logger:logging.Logger, host:str='localhost', port:int=6379, password:str='password', svname:str='server', maxrecsize:int=1000):
        """
        推論結果をshowimg.htmlに転送する後処理クラスです。
        
        Args:
            logger (logging.Logger): ロガー
            host (str, optional): ホスト名。デフォルトは'localhost'。
            port (int, optional): ポート番号。デフォルトは6379。
            password (str, optional): パスワード。デフォルトは'password'。
            svname (str, optional): サーバ名。デフォルトは'server'。
            maxrecsize (int, optional): 最大レコードサイズ。デフォルトは1000。
        """
        super().__init__(logger)
        self.config = dict(host=host, port=port, password=password, svname=svname, maxrecsize=maxrecsize)
        self.injection = after_showimg_injection.AfterShowimgInjection(self.config, self.logger)

    def post(self, outputs:Dict[str, Any], output_image:Image.Image) -> Tuple[Dict[str, Any], Image.Image]:
        """
        後処理を行う関数です。

        Args:
            outputs (Dict[str, Any]): 推論結果
            output_image (Image.Image): 入力画像（RGB配列であること）

        Returns:
            Dict[str, Any]: 後処理結果
            Image: 後処理結果
        """
        outputs, output_image = self.injection.action(None, None, outputs, output_image, None)
        return outputs, output_image



from PIL import Image
from typing import List, Tuple, Dict, Any
import logging

class Injection(object):
    def __init__(self, config:Dict[str,Any], logger:logging.Logger):
        """
        このクラスのインスタンスを初期化します。
        継承時は、このコンストラクタを呼び出すようにしてください。super().__init__(logger)

        Args:
            config (Dict[str,Any]): 設定
            logger (logging.Logger): ロガー
        Returns:
            BeforeInjection
        """
        self.config = config
        self.logger = logger

    def add_warning(self, outputs:Dict[str, Any], message:str):
        """
        警告メッセージを追加します。

        Args:
            outputs (Dict[str, Any]): 推論結果
            message (str): 警告メッセージ
        """
        if 'injection_warn' not in outputs:
            outputs['injection_warn'] = []
        self.logger.warning(f'({type(self)}): {message}')
        outputs['injection_warn'].append(f'({type(self)}): {message}')

    def add_success(self, outputs:Dict[str, Any], message:Any):
        """
        成功メッセージを追加します。

        Args:
            outputs (Dict[str, Any]): 推論結果
            message (Any): 成功メッセージ
        """
        if 'injection_success' not in outputs:
            outputs['injection_success'] = []
        outputs['injection_success'].append(message)

    def get_config(self, key:str, default:Any=None) -> Any:
        """
        設定を取得します。
        Args:
            key (str): キー
            default (Any): デフォルト値
        Returns:
            Any: 設定値
        """
        ret = self.config.get(key, default)
        return ret if ret is not None else default

    def _set_config(self, key:str, val:Any=None) -> Any:
        """
        設定を設定します。
        Args:
            key (str): キー
            val (Any): 値
        """
        self.config[key] = val

class BeforeInjection(Injection):
    """
    サーバーサイドで実行する前処理を実装するためのクラスです。
    """

    def action(self, reskey:str, name:str, image:Image.Image, session:Dict[str, Any]) -> Image.Image:
        """
        このメソッドは推論を実行する前処理を実行します。

        Args:
            reskey (str): レスポンスキー
            name (str): モデル名
            image (Image.Image): 推論する画像データ
            session (Dict[str, Any]): 推論セッション。次の項目が含まれます。
            ・session: app.predict.Predict#create_session() で作成されたセッション
            ・model_img_width: モデルの入力画像の幅
            ・model_img_height: モデルの入力画像の高さ
            ・predict_obj: app.predict.Predict インスタンス
            ・labels: クラスラベルのリスト
            ・colors: ボックスの色のリスト
            ・tracker: use_trackがTrueの場合、トラッカーのインスタンス
        Returns:
            Image.Image: 前処理後の画像データ
        """
        return image

class AfterInjection(Injection):
    """
    サーバーサイドで実行する後処理を実装するためのクラスです。
    """

    def action(self, reskey:str, name:str, outputs:Dict[str, Any], output_image:Image.Image, session:Dict[str, Any]) -> Tuple[Dict[str, Any], Image.Image]:
        """
        このメソッドは推論を実行した後の処理を実行します。

        Args:
            reskey (str): レスポンスキー
            name (str): モデル名
            outputs (Dict[str, Any]): 推論結果。次の項目が含まれます。
                ・success or warn: 推論成功か警告のキーに対して、その内容が格納されます。
                ・output_image: 推論後の画像データをbase64エンコードした文字列
                ・output_image_shape: 推論後の画像データの形状（base46でコードするときに必要）
                ・output_image_name: クライアントから指定されてきた推論後の画像データの名前

            output_image (Image.Image): 推論後の画像データ
            session (Dict[str, Any]): 推論セッション。次の項目が含まれます。
                ・session: app.predict.Predict#create_session() で作成されたセッション
                ・model_img_width: モデルの入力画像の幅
                ・model_img_height: モデルの入力画像の高さ
                ・predict_obj: app.predict.Predict インスタンス
                ・labels: クラスラベルのリスト
                ・colors: ボックスの色のリスト
                ・tracker: use_trackがTrueの場合、トラッカーのインスタンス

        Returns:
            Tuple[Dict[str, Any], Image.Image]: 後処理後の推論結果と画像データのタプル
        """
        return outputs, output_image



from cmdbox.app import common, feature
from cmdbox.app.commons import redis_client
from cmdbox.app.options import Options
from iinfer.app import client, common as cmn
from iinfer.app.commons import module
from pathlib import Path
from typing import Dict, Any, Tuple, Union, List
import argparse
import base64
import logging
import json
import shutil
import urllib

class ClientDeploy(feature.OneshotNotifyEdgeFeature):
    def get_mode(self) -> Union[str, List[str]]:
        """
        この機能のモードを返します

        Returns:
            Union[str, List[str]]: モード
        """
        return 'client'

    def get_cmd(self):
        """
        この機能のコマンドを返します

        Returns:
            str: コマンド
        """
        return 'deploy'
    
    def get_option(self):
        """
        この機能のオプションを返します

        Returns:
            Dict[str, Any]: オプション
        """
        return dict(
            use_redis=self.USE_REDIS_TRUE, nouse_webmode=False,
            description_ja="AIモデルをサーバーに配備します。",
            description_en="Deploy AI model to server.",
            choice=[
                dict(opt="host", type=Options.T_STR, default=self.default_host, required=True, multi=False, hide=True, choice=None, web="mask",
                     description_ja="Redisサーバーのサービスホストを指定します。",
                     description_en="Specify the service host of the Redis server.",
                     test_true={"yolox":"localhost"},
                     test_false={"yolox":"redis"}),
                dict(opt="port", type=Options.T_INT, default=self.default_port, required=True, multi=False, hide=True, choice=None, web="mask",
                     description_ja="Redisサーバーのサービスポートを指定します。",
                     description_en="Specify the service port of the Redis server.",
                     test_true={"yolox":6379},
                     test_false={"yolox":6380}),
                dict(opt="password", type=Options.T_STR, default=self.default_pass, required=True, multi=False, hide=True, choice=None, web="mask",
                     description_ja="Redisサーバーのアクセスパスワード(任意)を指定します。省略時は `password` を使用します。",
                     description_en="Specify the access password of the Redis server (optional). If omitted, `password` is used.",
                     test_true={"yolox":"password"},
                     test_false={"yolox":"password2"}),
                dict(opt="svname", type=Options.T_STR, default=self.default_svname, required=True, multi=False, hide=True, choice=None, web="readonly",
                     description_ja="サーバーのサービス名を指定します。省略時は `server` を使用します。",
                     description_en="Specify the service name of the inference server. If omitted, `server` is used.",
                     test_true={"yolox":"server"},
                     test_false={"yolox":"serverX"}),
                dict(short="n", opt="name", type=Options.T_STR, default=None, required=True, multi=False, hide=False, choice=None,
                     description_ja="AIモデルの登録名を指定します。",
                     description_en="Specify the registration name of the AI model.",
                     test_true={"yolox":"yolox",
                                "upernet":"upernet",
                                "san":"san",
                                "pspnet":"pspnet",
                                "swin":"swin",
                                "lnsightface":"lnsightface",
                                "yolo3":"yolo3",
                                "effnet":"effnet",
                                "custom":"custom"}),
                dict(opt="model_file", type=Options.T_FILE, default=None, required=True, multi=False, hide=False, choice=None, fileio="in",
                     description_ja="学習済みのモデルファイルのパス又はダウンロードURLを指定します。",
                     description_en="Specify the path or download URL of the trained model file.",
                     test_true={"yolox":"https://download.openmmlab.com/mmdetection/v2.0/yolox/yolox_s_8x8_300e_coco/yolox_s_8x8_300e_coco_20211121_095711-4592a793.pth",
                                "upernet":"https://download.openmmlab.com/mmsegmentation/v0.5/swin/upernet_swin_small_patch4_window7_512x512_160k_ade20k_pretrain_224x224_1K/upernet_swin_small_patch4_window7_512x512_160k_ade20k_pretrain_224x224_1K_20210526_192015-ee2fff1c.pth",
                                "san":"https://download.openmmlab.com/mmsegmentation/v0.5/san/san-vit-b16_20230906-fd0a7684.pth",
                                "pspnet":"https://download.openmmlab.com/mmsegmentation/v0.5/pspnet/pspnet_r18b-d8_512x1024_80k_cityscapes/pspnet_r18b-d8_512x1024_80k_cityscapes_20201226_063116-26928a60.pth",
                                "swin":"https://download.openmmlab.com/mmclassification/v0/swin-transformer/swin_small_224_b16x64_300e_imagenet_20210615_110219-7f9d988b.pth",
                                "lnsightface":"https://drive.usercontent.google.com/download?id=1pKIusApEfoHKDjeBTXYB3yOQ0EtTonNE&export=download&authuser=0&confirm=t&uuid=00c74cef-3534-49a3-942b-582771fad908&at=APZUnTXNi6MNLsiK-EMqx_cRMJ8a%3A1723645526732",
                                "yolo3":"https://github.com/onnx/models/raw/main/validated/vision/object_detection_segmentation/yolov3/model/yolov3-10.onnx",
                                "effnet":"https://github.com/onnx/models/raw/main/validated/vision/classification/efficientnet-lite4/model/efficientnet-lite4-11-qdq.onnx",
                                "custom":"https://download.openmmlab.com/mmdetection/v2.0/yolox/yolox_s_8x8_300e_coco/yolox_s_8x8_300e_coco_20211121_095711-4592a793.pth"},
                     test_false={"yolox":"https://download.openmmlab.com/mmdetection/v2.0/yolox/yolox_s_8x8_300e_coco/yolox_s_8x8_300e_coco_20211121_095711-4592a793.XXX.pth"}),
                dict(opt="model_conf_file", type=Options.T_FILE, default=None, required=False, multi=True, hide=False, choice=None, fileio="in",
                     description_ja="モデル設定ファイルを指定します。複数指定可能ですが、最初に指定したファイルが `start` 時に使用されます。",
                     description_en="Specify the model configuration file. Multiple specifications are possible, but the file specified first is used at `start` time.",
                     test_true={"yolox":["iinfer/extensions/configs/mmdet/yolox_s_8xb8-300e_coco.py","iinfer/extensions/configs/mmdet/yolox_tta.py"],
                                "upernet":["iinfer/extensions/configs/mmseg/swin-small-patch4-window7-in1k-pre_upernet_8xb2-160k_ade20k-512x512.py",
                                            "iinfer/extensions/configs/mmseg/swin-tiny-patch4-window7-in1k-pre_upernet_8xb2-160k_ade20k-512x512.py"],
                                "san":["iinfer/extensions/configs/mmseg/san-vit-b16_coco-stuff164k-640x640.py"],
                                "pspnet":["iinfer/extensions/configs/mmseg/pspnet_r18-d8_4xb2-80k_cityscapes-512x1024.py",
                                            "iinfer/extensions/configs/mmseg/pspnet_r50-d8_4xb2-80k_cityscapes-512x1024.py"],
                                "swin":["iinfer/extensions/configs/mmpretrain/swin-small_16xb64_in1k.py"],
                                "lnsightface":None,
                                "yolo3":None,
                                "effnet":None,
                                "custom":["iinfer/extensions/configs/mmdet/yolox_s_8xb8-300e_coco.py","iinfer/extensions/configs/mmdet/yolox_tta.py"]},
                     test_false={"yolox":["iinfer/extensions/configs/mmdet/yolox_s_8xb8-300e_coco.py","iinfer/extensions/configs/mmdet/yolox_tta.XX"]}),
                dict(opt="model_img_width", type=Options.T_INT, default=None, required=False, multi=False, hide=True, choice=None,
                     description_ja="AIモデルのINPUTサイズ(横px)を指定します。",
                     description_en="Specify the INPUT size (width px) of the AI model.",
                     test_true={"yolox":640,
                                "upernet":512,
                                "san":640,
                                "pspnet":512,
                                "swin":384,
                                "lnsightface":640,
                                "yolo3":416,
                                "effnet":224,
                                "custom":640}),
                dict(opt="model_img_height", type=Options.T_INT, default=None, required=False, multi=False, hide=True, choice=None,
                     description_ja="AIモデルのINPUTサイズ(縦px)を指定します。",
                     description_en="Specify the INPUT size (height px) of the AI model.",
                     test_true={"yolox":640,
                                "upernet":512,
                                "san":640,
                                "pspnet":512,
                                "swin":384,
                                "lnsightface":640,
                                "yolo3":416,
                                "effnet":224,
                                "custom":640}),
                dict(opt="predict_type", type=Options.T_STR, default=None, required=False, multi=False, hide=False,
                     choice=['','Custom']+[key for key in cmn.BASE_MODELS.keys()],
                     description_ja="AIモデルの推論タイプを指定します。",
                     description_en="Specify the inference type of the AI model.",
                     choice_show=dict(Custom=["custom_predict_py"]),
                     test_true={"yolox":"mmdet_det_YoloX",
                                "upernet":"mmseg_seg_SwinUpernet",
                                "san":"mmseg_seg_San",
                                "pspnet":"mmseg_seg_PSPNet",
                                "swin":"mmpretrain_cls_swin",
                                "lnsightface":"insightface_det",
                                "yolo3":"onnx_det_YoloV3",
                                "effnet":"onnx_cls_EfficientNet_Lite4",
                                "custom":"Custom"}),
                dict(opt="custom_predict_py", type=Options.T_FILE, default=None, required=False, multi=False, hide=False, choice=None, fileio="in",
                     description_ja="独自の推論タイプを作成するときに指定。この時は `--predict_type Custom` を指定。",
                     description_en="Specify when creating a custom inference type. In this case, specify `--predict_type Custom`.",
                     test_true={"yolox":None,
                                "custom":"iinfer/tools/datas/predicts/mmdet_det_YoloX2.py"}),
                dict(opt="label_file", type=Options.T_FILE, default=None, required=False, multi=False, hide=False, choice=None, fileio="in",
                     description_ja="推論結果のクラスラベルファイルを指定。改行区切りでラベル名(行indexがクラスと一致する)を指定したファイル。",
                     description_en="Specify the class label file of the inference result. A file specifying the label name (the row index matches the class) separated by line breaks.",
                     test_true={"yolox":"iinfer/extensions/label_coco.txt",
                                "upernet":"iinfer/extensions/label_imagenet1k.txt",
                                "san":"iinfer/extensions/label_imagenet1k.txt",
                                "pspnet":"iinfer/extensions/label_cityscapes.txt",
                                "swin":"iinfer/extensions/label_imagenet1k.txt",
                                "lnsightface":None,
                                "yolo3":None,
                                "effnet":None,
                                "custom":"iinfer/extensions/label_coco.txt"},
                     test_false={"yolox":"iinfer/extensions/label_coco.ttt"}),
                dict(opt="color_file", type=Options.T_FILE, default=None, required=False, multi=False, hide=False, choice=None, fileio="in",
                     description_ja="推論結果の可視化画像の色ファイルを指定。改行区切りで色(行indexがクラスと一致する)を指定したファイル。",
                     description_en="Specify the color file of the visualization image of the inference result. A file specifying the color (the row index matches the class) separated by line breaks.",
                     test_true={"yolox":None}),
                dict(opt="before_injection_type", type=Options.T_STR, default=None, required=False, multi=True, hide=True, choice=['']+[key for key in cmn.BASE_BREFORE_INJECTIONS.keys()],
                     description_ja="前処理を実行させるときに指定。",
                     description_en="Specify when you want to execute preprocessing.",
                     test_true={"yolox":[None, "before_grayimg_injection"]},
                     test_false={"yolox":["before_grayimg_injection2"]}),
                dict(opt="before_injection_conf", type=Options.T_FILE, default=None, required=False, multi=False, hide=True, choice=None, fileio="in",
                     description_ja="前処理に対する設定ファイルを指定。",
                     description_en="Specify the setting file for preprocessing.",
                     test_true={"yolox":"iinfer/tools/datas/injections/before_injection.json"},
                     test_false={"yolox":"iinfer/tools/datas/injections/before_injection.XXX"}),
                dict(opt="before_injection_py", type=Options.T_FILE, default=None, required=False, multi=True, hide=True, choice=None, fileio="in",
                     description_ja="独自の前処理を作成するときに指定。",
                     description_en="Specify when creating a custom preprocessing.",
                     test_true={"yolox":None,
                                "custom":"iinfer/tools/datas/injections/before_grayimg_injection2.py"},
                     test_false={"yolox":None,
                                 "custom":"iinfer/tools/datas/injections/before_grayimg_injection2.XXX"}),
                dict(opt="after_injection_type", type=Options.T_STR, default=None, required=False, multi=True, hide=True, choice=['']+[key for key in cmn.BASE_AFTER_INJECTIONS.keys()],
                     description_ja="後処理を作成させるときに指定。",
                     description_en="Specify when you want to create post-processing.",
                     test_true={"yolox":["after_det_filter_injection","after_det_judge_injection"],
                                "upernet":["after_seg_filter_injection","after_seg_bbox_injection","after_det_filter_injection"],
                                "san":["after_seg_filter_injection","after_seg_bbox_injection","after_det_filter_injection"],
                                "pspnet":["after_seg_filter_injection","after_seg_bbox_injection","after_det_filter_injection"],
                                "swin":["after_cls_judge_injection"],
                                "lnsightface":None,
                                "yolo3":["after_det_filter_injection","after_det_judge_injection"],
                                "effnet":["after_cls_judge_injection"],
                                "custom":["after_det_filter_injection","after_det_judge_injection"]},
                     test_false={"yolox":["after_det_filter_injection","after_det_judge_injection2"]}),
                dict(opt="after_injection_conf", type=Options.T_FILE, default=None, required=False, multi=False, hide=True, choice=None, fileio="in",
                     description_ja="後処理に対する設定ファイルを指定。",
                     description_en="Specify the setting file for post-processing.",
                     test_true={"yolox":"iinfer/tools/datas/injections/after_injection.json"},
                     test_false={"yolox":"iinfer/tools/datas/injections/after_injection.XXX"}),
                dict(opt="after_injection_py", type=Options.T_FILE, default=None, required=False, multi=True, hide=True, choice=None, fileio="in",
                     description_ja="独自の後処理を作成するときに指定。",
                     description_en="Specify when creating custom post-processing.",
                     test_true={"yolox":None,
                                "custom":"iinfer/tools/datas/injections/after_det_filter_injection2.py"},
                     test_false={"yolox":None,
                                    "custom":"iinfer/tools/datas/injections/after_det_filter_injection2.XXX"}),
                dict(opt="overwrite", type=Options.T_BOOL, default=True, required=False, multi=False, hide=False, choice=[True, False],
                     description_ja="デプロイ済みであっても上書きする指定。",
                     description_en="Specify to overwrite even if it is already deployed.",
                     test_true={"yolox":True}),
                dict(opt="train_type", type=Options.T_STR, default=None, required=False, multi=False, hide=False,
                     choice=['','Custom']+[key for key in cmn.BASE_TRAIN_MODELS.keys()],
                     description_ja="AIモデルの学習タイプを指定します。",
                     description_en="Specify the train type of the AI model.",
                     choice_show=dict(**{key:["train_dataset","train_dataset_upload"] for key in list(cmn.BASE_TRAIN_MODELS.keys())},
                                         Custom=["train_dataset","train_dataset_upload","custom_train_py"]),
                     test_true={"yolox":"mmdet_det_YoloX",
                                "upernet":"mmseg_seg_SwinUpernet",
                                "san":"mmseg_seg_San",
                                "pspnet":"mmseg_seg_PSPNet",
                                "swin":"mmpretrain_cls_swin",
                                "lnsightface":None,
                                "yolo3":None,
                                "effnet":None,
                                "custom":"Custom"}),
                dict(opt="train_dataset", type=Options.T_DIR, default=None, required=False, multi=False, hide=False, choice=None, fileio="in",
                     description_ja="データセットディレクトリを指定します。",
                     description_en="Specifies the data set directory.",
                     test_true={"yolox":"iinfer/extensions/data",
                                "lnsightface":None,
                                "yolo3":None,
                                "effnet":None}),
                dict(opt="train_dataset_upload", type=Options.T_BOOL, default=False, required=False, multi=False, hide=False, choice=[True, False],
                     description_ja="データセットをサーバーにアップロードします。",
                     description_en="Upload the data set to the server.",
                     test_true={"yolox":True,
                                "lnsightface":False,
                                "yolo3":False,
                                "effnet":False}),
                dict(opt="custom_train_py", type=Options.T_FILE, default=None, required=False, multi=False, hide=False, choice=None, fileio="in",
                     description_ja="独自の学習タイプを作成するときに指定。この時は `--train_type Custom` を指定。",
                     description_en="Specify when creating a custom train type. In this case, specify `--train_type Custom`.",
                     test_true={"yolox":None,
                                "custom":"iinfer/tools/datas/trains/mmdet_det_YoloX2.py"}),
                dict(opt="retry_count", type=Options.T_INT, default=3, required=False, multi=False, hide=True, choice=None,
                     description_ja="Redisサーバーへの再接続回数を指定します。0以下を指定すると永遠に再接続を行います。",
                     description_en="Specifies the number of reconnections to the Redis server.If less than 0 is specified, reconnection is forever.",
                     test_true={"yolox":1}),
                dict(opt="retry_interval", type=Options.T_INT, default=5, required=False, multi=False, hide=True, choice=None,
                     description_ja="Redisサーバーに再接続までの秒数を指定します。",
                     description_en="Specifies the number of seconds before reconnecting to the Redis server.",
                     test_true={"yolox":1}),
                dict(opt="timeout", type=Options.T_INT, default=120, required=False, multi=False, hide=False, choice=None,
                     description_ja="サーバーの応答が返ってくるまでの最大待ち時間を指定。",
                     description_en="Specify the maximum waiting time until the server responds.",
                     test_true={"yolox":60}),
                dict(opt="output_json", short="o", type=Options.T_FILE, default=None, required=False, multi=False, hide=True, choice=None, fileio="out",
                     description_ja="処理結果jsonの保存先ファイルを指定。",
                     description_en="Specify the destination file for saving the processing result json.",
                     test_true={"yolox":None,
                                "custom":"pred.json"}),
                dict(opt="output_json_append", short="a", type=Options.T_BOOL, default=False, required=False, multi=False, hide=True, choice=[True, False],
                     description_ja="処理結果jsonファイルを追記保存します。",
                     description_en="Save the processing result json file by appending.",
                     test_true={"yolox":False}),
                dict(opt="stdout_log", type=Options.T_BOOL, default=True, required=False, multi=False, hide=True, choice=[True, False],
                     description_ja="GUIモードでのみ使用可能です。コマンド実行時の標準出力をConsole logに出力します。",
                     description_en="Available only in GUI mode. Outputs standard output during command execution to Console log."),
                dict(opt="capture_stdout", type=Options.T_BOOL, default=True, required=False, multi=False, hide=True, choice=[True, False],
                     description_ja="GUIモードでのみ使用可能です。コマンド実行時の標準出力をキャプチャーし、実行結果画面に表示します。",
                     description_en="Available only in GUI mode. Captures standard output during command execution and displays it on the execution result screen."),
                dict(opt="capture_maxsize", type=Options.T_INT, default=self.DEFAULT_CAPTURE_MAXSIZE, required=False, multi=False, hide=True, choice=None,
                     description_ja="GUIモードでのみ使用可能です。コマンド実行時の標準出力の最大キャプチャーサイズを指定します。",
                     description_en="Available only in GUI mode. Specifies the maximum capture size of standard output when executing commands."),
            ]
        )

    def get_svcmd(self):
        """
        この機能のサーバー側のコマンドを返します

        Returns:
            str: サーバー側のコマンド
        """
        return 'deploy'

    def apprun(self, logger:logging.Logger, args:argparse.Namespace, tm:float, pf:List[Dict[str, float]]=[]) -> Tuple[int, Dict[str, Any], Any]:
        """
        この機能の実行を行います

        Args:
            logger (logging.Logger): ロガー
            args (argparse.Namespace): 引数
            tm (float): 実行開始時間
            pf (List[Dict[str, float]]): 呼出元のパフォーマンス情報

        Returns:
            Tuple[int, Dict[str, Any], Any]: 終了コード, 結果, オブジェクト
        """
        if args.svname is None:
            msg = {"warn":f"Please specify the --svname option."}
            common.print_format(msg, args.format, tm, args.output_json, args.output_json_append, pf=pf)
            return self.RESP_WARN, msg, None
        cl = client.Client(logger, redis_host=args.host, redis_port=args.port, redis_password=args.password, svname=args.svname)

        if args.model_conf_file is not None:
            args.model_conf_file = [Path(f) for f in args.model_conf_file if f is not None and f != '']
        if args.before_injection_py is not None:
            args.before_injection_py = [Path(f) for f in args.before_injection_py if f is not None and f != '']
        if args.after_injection_py is not None:
            args.after_injection_py = [Path(f) for f in args.after_injection_py if f is not None and f != '']
        args.custom_predict_py = Path(args.custom_predict_py) if args.custom_predict_py is not None else None
        args.label_file = Path(args.label_file) if args.label_file is not None else None
        args.color_file = Path(args.color_file) if args.color_file is not None else None
        args.before_injection_conf = Path(args.before_injection_conf) if args.before_injection_conf is not None else None
        args.after_injection_conf = Path(args.after_injection_conf) if args.after_injection_conf is not None else None

        args.train_dataset = Path(args.train_dataset) if args.train_dataset is not None else None
        args.custom_train_py = Path(args.custom_train_py) if args.custom_train_py is not None else None

        ret = cl.deploy(args.name, args.model_img_width, args.model_img_height, args.model_file, args.model_conf_file, args.predict_type,
                        args.custom_predict_py, label_file=args.label_file, color_file=args.color_file,
                        before_injection_conf=args.before_injection_conf, before_injection_type=args.before_injection_type, before_injection_py=args.before_injection_py,
                        after_injection_conf=args.after_injection_conf, after_injection_type=args.after_injection_type, after_injection_py=args.after_injection_py,
                        train_dataset=args.train_dataset, train_dataset_upload=args.train_dataset_upload, train_type=args.train_type, custom_train_py=args.custom_train_py,
                        overwrite=args.overwrite, retry_count=args.retry_count, retry_interval=args.retry_interval, timeout=args.timeout)
        common.print_format(ret, args.format, tm, args.output_json, args.output_json_append, pf=pf)
        if 'success' not in ret:
            return self.RESP_WARN, ret, cl

        return self.RESP_SUCCESS, ret, cl

    def is_cluster_redirect(self):
        """
        クラスター宛のメッセージの場合、メッセージを転送するかどうかを返します

        Returns:
            bool: メッセージを転送する場合はTrue
        """
        return True

    def svrun(self, data_dir:Path, logger:logging.Logger, redis_cli:redis_client.RedisClient, msg:List[str],
              sessions:Dict[str, Dict[str, Any]]) -> int:
        """
        この機能のサーバー側の実行を行います

        Args:
            data_dir (Path): データディレクトリ
            logger (logging.Logger): ロガー
            redis_cli (redis_client.RedisClient): Redisクライアント
            msg (List[str]): 受信メッセージ
            sessions (Dict[str, Dict[str, Any]]): セッション情報
        
        Returns:
            int: 終了コード
        """

        if msg[7] == 'None':
            model_bin = None
        else:
            model_bin = base64.b64decode(msg[7])
        if msg[8] == 'None':
            model_conf_file = None
        else:
            model_conf_file = msg[8].split(',')
        if msg[9] == 'None':
            model_conf_bin = None
        else:
            model_conf_bin = [base64.b64decode(m) for m in msg[9].split(',')]
        if msg[10] == 'None':
            custom_predict_py = None
        else:
            custom_predict_py = base64.b64decode(msg[10])
        if msg[11] == 'None':
            label_txt = None
        else:
            label_txt = base64.b64decode(msg[11])
        if msg[12] == 'None':
            color_txt = None
        else:
            color_txt = base64.b64decode(msg[12])
        if msg[13] == 'None':
            before_injection_conf = None
        else:
            before_injection_conf = base64.b64decode(msg[13])
        if msg[14] == 'None':
            before_injection_type = None
        else:
            before_injection_type = msg[14].split(',')
        if msg[15] == 'None':
            before_injection_py = None
        else:
            before_injection_py = msg[15].split(',')
        if msg[16] == 'None':
            before_injection_bin = None
        else:
            before_injection_bin = [base64.b64decode(m) for m in msg[16].split(',')]
        if msg[17] == 'None':
            after_injection_conf = None
        else:
            after_injection_conf = base64.b64decode(msg[17])
        if msg[18] == 'None':
            after_injection_type = None
        else:
            after_injection_type = msg[18].split(',')
        if msg[19] == 'None':
            after_injection_py = None
        else:
            after_injection_py = msg[19].split(',')
        if msg[20] == 'None':
            after_injection_bin = None
        else:
            after_injection_bin = [base64.b64decode(m) for m in msg[20].split(',')]
        if msg[21] == 'None':
            train_dataset = None
        else:
            train_dataset = msg[21]
        if msg[22] == 'None':
            train_type = None
        else:
            train_type = msg[22]
        if msg[23] == 'None':
            custom_train_py = None
        else:
            custom_train_py = base64.b64decode(msg[23])
        if msg[24] == 'True':
            overwrite = True
        else:
            overwrite = False

        st = self.deploy(msg[1], msg[2], int(msg[3]), int(msg[4]), msg[5], msg[6],
                            model_bin, model_conf_file, model_conf_bin, custom_predict_py, label_txt, color_txt,
                            before_injection_conf, before_injection_type, before_injection_py, before_injection_bin,
                            after_injection_conf, after_injection_type, after_injection_py, after_injection_bin,
                            train_dataset, train_type, custom_train_py, overwrite,
                            data_dir, logger, redis_cli, sessions)

        return st

    def deploy(self, reskey:str, name:str, model_img_width:int, model_img_height:int, predict_type:str,
               model_file:str, model_bin:bytes, model_conf_file:List[str], model_conf_bin:List[bytes],
               custom_predict_py:bytes, label_txt:bytes, color_txt:bytes,
               before_injection_conf:bytes, before_injection_type:List[str], before_injection_py:List[str], before_injection_bin:List[bytes],
               after_injection_conf:bytes, after_injection_type:List[str], after_injection_py:List[str], after_injection_bin:List[bytes],
               train_dataset:str, train_type:str, custom_train_py:bytes, overwrite:bool,
               data_dir:Path, logger:logging.Logger, redis_cli:redis_client.RedisClient, sessions:Dict[str, Dict[str, Any]]):
        """
        モデルをデプロイする

        Args:
            reskey (str): レスポンスキー
            name (dict): モデル名
            model_img_width (int): 画像の幅
            model_img_height (int): 画像の高さ
            predict_type (str): 推論方法のタイプ
            model_file (str): モデルのファイル名
            model_bin (bytes): モデルファイル
            model_conf_file (List[str]): モデル設定のファイル名
            model_conf_bin (List[bytes]): モデル設定ファイル
            custom_predict_py (bytes): 推論のPythonスクリプト
            label_txt (bytes): ラベルファイル
            color_txt (bytes): 色設定ファイル
            before_injection_conf (bytes): 推論前処理の設定ファイル
            before_injection_type (List[str]): 推論前処理のタイプ
            before_injection_py (List[str]): 推論前処理のPythonスクリプトファイル名
            before_injection_bin (List[bytes]): 推論前処理のPythonスクリプト
            after_injection_conf (bytes): 推論後処理の設定ファイル
            after_injection_type (List[str]): 推論後処理のタイプ
            after_injection_py (List[str]): 推論後処理のPythonスクリプトファイル名
            after_injection_bin (List[bytes]): 推論後処理のPythonスクリプト
            train_dataset (str): 学習データセットのパス
            train_type (str): 学習方法のタイプ
            custom_train_py (bytes): 学習のPythonスクリプト
            overwrite (bool): 上書きするかどうか
            data_dir (Path): データディレクトリ
            logger (logging.Logger): ロガー
            redis_cli (redis_client.RedisClient): Redisクライアント
            sessions (Dict[str, Dict[str, Any]]): セッション情報
        """
        if name is None or name == "":
            logger.warning(f"Name is empty.")
            redis_cli.rpush(reskey, {"warn": f"Name is empty."})
            return self.RESP_WARN
        if model_file is None or model_file == "":
            logger.warning(f"model_file is empty.")
            redis_cli.rpush(reskey, {"warn": f"model_file is empty."})
            return self.RESP_WARN
        if model_img_width is None or model_img_width <= 0:
            logger.warning(f"Image width is invalid.")
            redis_cli.rpush(reskey, {"warn": f"Image width is invalid."})
            return self.RESP_WARN
        if model_img_height is None or model_img_height <= 0:
            logger.warning(f"Image height is invalid.")
            redis_cli.rpush(reskey, {"warn": f"Image height is invalid."})
            return self.RESP_WARN

        deploy_dir = data_dir / name
        if name in sessions:
            logger.warning(f"{name} has already started a session.")
            redis_cli.rpush(reskey, {"warn": f"{name} has already started a session."})
            return self.RESP_WARN
        if not overwrite and deploy_dir.exists():
            logger.warning(f"Could not be deployed. '{deploy_dir}' already exists")
            redis_cli.rpush(reskey, {"warn": f"Could not be deployed. '{deploy_dir}' already exists"})
            return self.RESP_WARN

        if predict_type != "Custom":
            if predict_type not in cmn.BASE_MODELS:
                logger.warning(f"Incorrect predict_type. '{predict_type}'")
                redis_cli.rpush(reskey, {"warn": f"Incorrect predict_type. '{predict_type}'"})
                return self.RESP_WARN
            if cmn.BASE_MODELS[predict_type]['required_model_conf']==True and model_conf_file is None:
                logger.warning(f"model_conf_file is None.")
                redis_cli.rpush(reskey, {"warn": f"model_conf_file is None."})
                return self.RESP_WARN

        common.mkdirs(deploy_dir)
        def _save_s(logger:logging.Logger, file:str, data:bytes, ret_fn:bool=False):
            if file is None or data is None:
                return False, None
            file = deploy_dir / file
            with open(file, "wb") as f:
                f.write(data)
                logger.info(f"Save {file} to {str(deploy_dir)}")
            return True, file if ret_fn else None

        if model_file.startswith("http") and (model_bin is None or model_bin == ''):
            model_path = deploy_dir / urllib.parse.urlparse(model_file).path.split('/')[-1]
            if not model_path.exists():
                logger.info(f"Downloading. {model_file}")
                urllib.request.urlretrieve(model_file, model_path)
                logger.info(f"Save {model_path}")
            else:
                logger.info(f"Already exists. {model_path}")
            model_file = model_path
        else:
            ret, model_file = _save_s(logger, model_file, model_bin, ret_fn=True)

        ret, before_injection_conf = _save_s(logger, "before_injection_conf.json", before_injection_conf, ret_fn=True)
        ret, after_injection_conf = _save_s(logger, "after_injection_conf.json", after_injection_conf, ret_fn=True)

        def _save_m(logger:logging.Logger, name:str, files:List[str], datas:List[bytes]):
            if files is not None and datas is None:
                logger.warning(f"{name}_file is not None but {name}_bin is None.")
                redis_cli.rpush(reskey, {"warn": f"{name}_file is not None but {name}_bin is None."})
                return False, files
            if files is None and datas is not None:
                logger.warning(f"{name}_file is None but {name}_bin is not None.")
                redis_cli.rpush(reskey, {"warn": f"{name}_file is None but {name}_bin is not None."})
                return False, files
            if files is not None:
                files = [deploy_dir / cf for cf in files if cf is not None and cf != '']
                for i, cf in enumerate(files):
                    with open(cf, "wb") as f:
                        f.write(datas[i])
                        logger.info(f"Save {cf} to {str(deploy_dir)}")
            return True, files
        ret, model_conf_file = _save_m(logger, 'model_conf', model_conf_file, model_conf_bin)
        if not ret: return self.RESP_WARN
        ret, before_injection_py = _save_m(logger, 'before_injection', before_injection_py, before_injection_bin)
        if not ret: return self.RESP_WARN
        ret, after_injection_py = _save_m(logger, 'after_injection', after_injection_py, after_injection_bin)
        if not ret: return self.RESP_WARN
        ret, custom_predict_file = _save_s(logger, 'custom_predict.py', custom_predict_py, ret_fn=True)
        ret, label_file = _save_s(logger, 'label.txt', label_txt, ret_fn=True)
        ret, color_file = _save_s(logger, 'color.txt', color_txt, ret_fn=True)

        if train_dataset is not None:
            train_dataset = deploy_dir / train_dataset
        else:
            train_dataset = None
        ret, custom_train_file = _save_s(logger, "custom_train.py", custom_train_py, ret_fn=True)

        with open(deploy_dir / "conf.json", "w") as f:
            conf = dict(appid=self.ver.__appid__, model_img_width=model_img_width, model_img_height=model_img_height, predict_type=predict_type,
                        model_file=model_file, model_conf_file=model_conf_file, custom_predict_py=(custom_predict_file if custom_predict_file is not None else None),
                        label_file=label_file, color_file=color_file, before_injection_conf=before_injection_conf, after_injection_conf=after_injection_conf,
                        before_injection_type=before_injection_type, after_injection_type=after_injection_type,
                        before_injection_py=before_injection_py, after_injection_py=after_injection_py,
                        train_dataset=train_dataset, train_type=train_type, custom_train_py=(custom_train_file if custom_train_file is not None else None),
                        deploy_dir=deploy_dir)
            json.dump(conf, f, default=common.default_json_enc, indent=4)
            logger.info(f"Save conf.json to {str(deploy_dir)}")

        self._gitpull(data_dir, logger, redis_cli, reskey, deploy_dir, predict_type)

        try:
            ret, predict_obj = module.build_predict(conf["predict_type"], conf["custom_predict_py"], logger)
            if not ret:
                redis_cli.rpush(reskey, predict_obj)
                return self.RESP_WARN
            predict_obj.post_deploy(deploy_dir, conf)
        except Exception as e:
            logger.warning(f"Failed to load Predict: {e}", exc_info=True)
            redis_cli.rpush(reskey, {"warn": f"Failed to load Predict: {e}"})
            return self.RESP_WARN

        redis_cli.rpush(reskey, {"success": f"Save conf.json to {str(deploy_dir)}"})
        return self.RESP_SUCCESS

    def _gitpull(self, data_dir:Path, logger:logging.Logger, redis_cli:redis_client.RedisClient, reskey:str, deploy_dir:Path, predict_type:str):
        if predict_type.startswith('mmpretrain_'):
            if not (data_dir / "mmpretrain").exists():
                returncode, _, _cmd = common.cmd(f'cd {data_dir} && git clone https://github.com/open-mmlab/mmpretrain.git', logger=logger)
                if returncode != 0:
                    logger.warning(f"Failed to git clone mmpretrain. cmd: {_cmd}")
                    redis_cli.rpush(reskey, {"error": f"Failed to git clone mmpretrain. cmd: {_cmd}"})
                    return self.RESP_ERROR
            shutil.copytree(data_dir / "mmpretrain" / "configs", deploy_dir / "configs", dirs_exist_ok=True)
            logger.info(f"Copy mmpretrain configs to {str(deploy_dir / 'configs')}")
        elif predict_type.startswith('mmdet_'):
            if not (data_dir / "mmdetection").exists():
                returncode, _, _cmd = common.cmd(f'cd {data_dir} && git clone https://github.com/open-mmlab/mmdetection.git', logger=logger)
                if returncode != 0:
                    logger.warning(f"Failed to git clone mmdetection. cmd: {_cmd}")
                    redis_cli.rpush(reskey, {"error": f"Failed to git clone mmdetection. cmd: {_cmd}"})
                    return self.RESP_ERROR
            shutil.copytree(data_dir / "mmdetection" / "configs", deploy_dir / "configs", dirs_exist_ok=True)
            logger.info(f"Copy mmdetection configs to {str(deploy_dir / 'configs')}")
        elif predict_type.startswith('mmseg_'):
            if not (data_dir / "mmsegmentation").exists():
                returncode, _, _cmd = common.cmd(f'cd {data_dir} && git clone -b main https://github.com/open-mmlab/mmsegmentation.git', logger=logger)
                if returncode != 0:
                    logger.warning(f"Failed to git clone mmsegmentation. cmd: {_cmd}")
                    redis_cli.rpush(reskey, {"error": f"Failed to git clone mmsegmentation. cmd: {_cmd}"})
                    return self.RESP_ERROR
            shutil.copytree(data_dir / "mmsegmentation" / "configs", deploy_dir / "configs", dirs_exist_ok=True)
            logger.info(f"Copy mmsegmentation configs to {str(deploy_dir / 'configs')}")
